# -*- coding: utf-8 -*- 
# logging - provides primitive logging with logging levels
#
# author: Dr. Thomas Tensi, 2014-04

#====================

import sys
import time

from basemodules.simpletypes import Boolean, Natural, String
from basemodules.ttbase import adaptToRange, iif, isMicroPython
from basemodules.typesupport import toUnicodeString

if not isMicroPython:
    import atexit
    import io
    import threading

#====================

class Logging_Level:
    """Defines the levels of logging"""

    # no logging
    noLogging           = 0
    # only logging of errors and assertion failures
    error               = 1
    # logging of errors and abbreviated exit-entry traces
    standardAbbreviated = 2
    # logging of errors and full exit-entry traces
    standard            = 3
    # full logging (including internal traces)
    verbose             = 4

    #--------------------

    @classmethod
    def fromString (cls,
                    st : String) -> String:
        """Converts string <st> to log level, returns noLogging when
           string cannot be converted"""

        result = cls.noLogging
        st = st.lower()
           
        if st == "verbose":
            result = cls.verbose
        elif st == "standard":
            result = cls.standard
        elif st == "standardabbreviated":
            result = cls.standardAbbreviated
        elif st == "error":
            result = cls.error
        else:
            result = cls.noLogging
       
        return result
    
    #--------------------

    @classmethod
    def toString (cls,
                  level : Natural) -> String:
        """Converts <level> to string"""

        if level == cls.verbose:
            result = "verbose"
        elif level == cls.standard:
            result = "standard"
        elif level == cls.standardAbbreviated:
            result = "standardAbbreviated"
        elif level == cls.error:
            result = "error"
        else:
            result = "noLogging"

        return result
        
#====================

class Logging:
    """Provides some primitive logging."""

   
    _referenceLevel             = Logging_Level.noLogging
    _fileName                   = ""
    _fileIsOpen                 = None
    _fileIsKeptOpen             = None
    _file                       = None
    _isActive                   = True
    _timeIsLogged               = True
    _timeFactor                 = 1
    _timeFractionalPartTemplate = ""
    _timeFractionalDigitCount   = 0
    _previousTimestamp          = 0
    _timeOfDayString            = ""
    _threadIDIsLogged           = False

    # buffer logs data before log file is opened, otherwise a
    # write-through will be done
    _buffer = []

    # the list of function names to be ignored when traversing
    # run-time stack for relevant function names */
    _ignoredFunctionNameList = \
        ("check", "_internalCheck", "log", "post", "pre",
         "trace", "traceError", "_traceWithLevel")

    #-- TRACE PREFICES --
    # length of allowed prefixes for template in a trace call
    _tracePrefixLength = 2

    # trace prefix used for internal traces within a function
    _innerTracePrefix = "--"

    # list of allowed prefixes for template in a entry-exit trace
    # call
    _entryExitPrefixList = (">>", "<<")
  
    # list of allowed prefixes for template in a trace call
    _standardPrefixList = _entryExitPrefixList + (_innerTracePrefix,)

    # --------------------
    # LOCAL FEATURES
    # --------------------

    @classmethod
    def _callingFunctionName (cls) -> String:
        """Returns function name of calling function.  Some functions
        are filtered out (like those from UI) and the class name is
        prepended."""

        if isMicroPython:
            return "..."
        else:
            callerDepth = 1
            found = False

            while not found:
                currentFrame = sys._getframe(callerDepth)
                codeObject = currentFrame.f_code
                functionName = codeObject.co_name
                found = (functionName not in cls._ignoredFunctionNameList)

                if not found:
                    callerDepth = callerDepth + 1
                else:
                    functionName = codeObject.co_qualname

        return functionName

    #--------------------

    @classmethod
    def _closeFileConditionally (cls):
        if cls._fileIsOpen:
            cls._file.close()
            cls._fileIsOpen = False
        
    #--------------------

    @classmethod
    def _currentTimeOfDay (cls) -> String:
        """Returns current time of day in seconds as string"""

        currentTimestamp = time.time()
        
        if currentTimestamp != cls._previousTimestamp:
            cls._previousTimestamp = currentTimestamp

            if not isMicroPython:
                st = time.strftime("%H%M%S")
            else:
                _, _, _, hours, minutes, seconds, _, _ = time.localtime()
                st = "%02d%02d%02d" % (hours, minutes, seconds)

            if cls._timeFractionalDigitCount > 0:
                fractionalPart = currentTimestamp - int(currentTimestamp)
                fractionalPart *= cls._timeFactor
                st += cls._timeFractionalPartTemplate % fractionalPart
                
            cls._timeOfDayString = st

        return cls._timeOfDayString
        
    #--------------------

    @classmethod
    def _openOrCreateFile (cls,
                           isNew : Boolean):
        """Creates or reopens logging file depending on value of
           <isNew>"""

        if cls._fileName == "":
            cls._file = None
        elif cls._fileName.lower() == "stderr":
            cls._file = sys.stderr
        else:
            mode = iif(isNew, "wt", "at")
            cls._file = io.open(cls._fileName, mode,
                                encoding="utf-8", errors='replace')

        cls._fileIsOpen = (cls._file is not None)
  
    #--------------------

    @classmethod
    def _prefixBefore (cls, st, otherSt):
        """Returns part of <st> before <otherSt>; if <otherSt> is not
           in string, the whole string will be returned"""

        splitPosition = st.indexOf(otherSt)
        result = st if splitPosition is None else st[0:splitPosition]
        return result
        
    #--------------------

    @classmethod
    def _traceWithLevel (cls,
                         level : Natural,
                         template : String,
                         *argumentList):
        """Writes <argumentList> formatted by <template> together with
           function name to log file."""

        if cls._isActive and level <= cls._referenceLevel:
            functionName = cls._callingFunctionName()
            prefixLength = cls._tracePrefixLength

            if template[0:prefixLength] not in cls._standardPrefixList:
                template = (cls._innerTracePrefix
                            + iif(len(template) > 0, ":", "")
                            + template)

            if cls._threadIDIsLogged:
                threadID = threading.current_thread().ident
                threadIDString = " [%d]" % threadID
            else:
                threadIDString = ""
                
            if cls._timeIsLogged:
                timeString = " (" + cls._currentTimeOfDay() + ")"
            else:
                timeString = ""

            st = (template[0:prefixLength]
                  + functionName
                  + timeString
                  + threadIDString)
            template = template[prefixLength:]

            # workaround for JYTHON or for mismatch in template and
            # argument list
            try:
                st += template % argumentList
            except:
                st += template + " ###CONVERSION ERROR###"

            if (cls._referenceLevel == Logging_Level.standardAbbreviated
                and st[0:prefixLength] in cls._standardPrefixList):
                # this is the level where the entry-exit data after
                # and including the colon is stripped off
                st = cls._prefixBefore(st, ":")
            
            cls._writeLine(st)

    #--------------------

    @classmethod
    def _writeLine (cls,
                    st : String):
        """Reopens logging file and writes single line <st>"""

        if isMicroPython:
            print(st)
        else:
            for pattern in ("\r\n", "\n", "\r"):
                st = st.replace(pattern, "<BR/>")

            st += '\n'

            if cls._fileName == "":
                # no output file => put line to buffer
                cls._buffer.append(st)
            else:
                if not cls._fileIsKeptOpen:
                    cls._openOrCreateFile(False)

                if cls._file is None:
                    # output file cannot be accessed => put line to buffer
                    cls._buffer.append(st)
                else:
                    cls._writeStringDirectly(st)

                if not cls._fileIsKeptOpen:
                    cls._file.close()

    #--------------------

    @classmethod
    def _writeStringDirectly (cls,
                              st : String):
        """Writes <st> to logging file"""

        st = toUnicodeString(st)
        cls._file.write(st)

    #--------------------
    # EXPORTED FEATURES
    #--------------------

    @classmethod
    def initialize (cls):
        """Starts logging"""

        cls._referenceLevel = Logging_Level.noLogging
        cls._fileName       = ""
        cls._fileIsOpen     = False
        cls._isActive       = True
        cls._buffer.clear()

        header = "START LOGGING -*- coding:utf-8 -*-"
        cls._writeLine(header)

        if not isMicroPython:
            atexit.register(cls._closeFileConditionally)
    
    #--------------------

    @classmethod
    def finalize (cls):
        """Ends logging."""

        footer = "END LOGGING"
        cls._writeLine(footer)

        if not isMicroPython:
            cls._closeFileConditionally()

    #--------------------
    #--------------------

    @classmethod
    def setEnabled (cls,
                    isEnabled : Boolean):
        """Sets logging to active or inactive"""

        cls._isActive = isEnabled

    #--------------------

    @classmethod
    def setLevel (cls,
                  loggingLevel : Logging_Level):
        """Sets logging reference level to <loggingLevel>"""

        cls._referenceLevel = loggingLevel

    #--------------------

    @classmethod
    def setFileName (cls,
                     fileName : String,
                     isKeptOpen : Boolean = True):
        """Sets file name for logging to <fileName>; if <isKeptOpen>
           is set, the logging file is not closed after each log entry"""

        if isMicroPython:
            cls._fileName = ""
        elif cls._fileName == fileName:
            cls._writeLine("logging file %s already open => skip"
                           % fileName)
        else:
            cls._fileName       = fileName
            cls._fileIsKeptOpen = isKeptOpen
            cls._openOrCreateFile(True)

            if cls._file is None:
                cls._fileName = ""
            else:
                for line in cls._buffer:
                    cls._writeStringDirectly(line)

                cls._buffer.clear()

            if not cls._fileIsKeptOpen:
                cls._file.close()

    #--------------------

    @classmethod
    def setTracingOfThreadID (cls,
                              threadIDIsLogged : Boolean):
        """Sets logging of threads to active or inactive depending on
           <threadIDIsLogged>"""

        Logging.trace(">>: %s", threadIDIsLogged)

        if isMicroPython:
            Logging.trace("--: multiprocessing not supported in MicroPython")
        else:
            cls._threadIDIsLogged = threadIDIsLogged

        Logging.trace("<<")

    #--------------------

    @classmethod
    def setTracingWithTime (cls,
                            timeIsLogged : Boolean,
                            fractionalDigitCount : Natural = 0):
        """Sets logging of time when tracing to active or inactive;
           <fractionalDigitCount> gives the number of fractional
           digits for the time logged"""

        fractionalDigitCount = adaptToRange(fractionalDigitCount, 0, 3)

        cls._timeIsLogged               = timeIsLogged
        cls._timeFractionalDigitCount   = fractionalDigitCount
        cls._timeFactor                 = 10 ** fractionalDigitCount
        cls._timeFractionalPartTemplate = \
            ".%0" + "%1d" % fractionalDigitCount + "d"

    #--------------------

    @classmethod
    def log (cls,
             st : String,
             level : Natural = Logging_Level.standard):
        """Writes <st> as a line to log file, when <level> is below or
           equal to the reference level."""

        if cls._referenceLevel >= level and cls._isActive:
            # log message is significant 
            cls._writeLine(st)

    #--------------------

    @classmethod
    def trace (cls,
               template : String,
               *argumentList):
        """Writes <argumentList> formatted by <template> together with
           function name to log file."""

        templatePrefix = template[0:cls._tracePrefixLength]
        isEntryExitTrace = templatePrefix in cls._entryExitPrefixList
        logLevel = iif(isEntryExitTrace,
                       Logging_Level.standardAbbreviated,
                       Logging_Level.verbose)
        cls._traceWithLevel(logLevel, template, *argumentList)

    #--------------------

    @classmethod
    def traceError (cls,
                    template : String,
                    *argumentList):
        """Writes <argumentList> formatted by <template> together with
           function name to log file as an error entry."""

        cls._traceWithLevel(Logging_Level.error,
                            "--: ERROR - " + template,
                            *argumentList)



# dispatcher for the soundfonttools programs adapting the
# module/packages paths accordingly

import os
import sys

# redirect sys.path to current project package directory and also
# include "_main" package
import soundfonttools
_currentModulePath = soundfonttools.__path__[0] + "/src"
sys.path.append(_currentModulePath)
sys.path.append(_currentModulePath + "/_main")

#--------------------

import _main
import _main.jsontosoundfont
import _main.soundfontanalyser
import _main.soundfontinplacerenamer
import _main.soundfonttojson

#--------------------

def _writeMessage (st):
    sys.stderr.write(st + "\n")

#--------------------

def main ():
    """Dispatch on name of calling program"""

    programPath = sys.argv[0]
    programName = os.path.basename(programPath).lower()
    # _writeMessage("program name = %s" % programName)

    if programName.startswith("jsontosoundfont"):
        _main.jsontosoundfont.main()
    elif programName.startswith("soundfonttojson"):
        _main.soundfonttojson.main()
    elif programName.startswith("soundfontinplacerenamer"):
        _main.soundfontinplacerenamer.main()
    elif programName.startswith("soundfontanalyser"):
        _main.soundfontanalyser.main()
    else:
        _writeMessage("unknown program name: %s" % programName)



# soundFontToJSON - scans SoundFont file and constructs JSON text file
#
# author: Dr. Thomas Tensi
# version: 2025-08

#====================
# IMPORTS
#====================

import argparse
import json
import sys

from basemodules.operatingsystem import OperatingSystem
from basemodules.simpleassertion import Assertion
from basemodules.simplelogging import Logging, Logging_Level
from basemodules.simpletypes import \
    Boolean, Natural, Object, ObjectList, String, StringList
from basemodules.ttbase import iif2
from basemodules.validitychecker import ValidityChecker

from multimedia.audio.wavefile import WaveFile
from multimedia.midi.soundfontfile import SoundFontFileReader
from multimedia.midi.soundfont import \
    SoundFont, SoundFontName, SoundFontSample, SoundFontSampleKind, \
    SoundFontWaveData

#====================

_programName = "soundFontToJSON"
_newline = "\n"

#--------------------
# error messages
#--------------------

_ErrMsg_duplicateSampleName = ("wave file export not possible:"
                               + " duplicate name '%s'")

#--------------------

def _findDuplicateNames (nameList : StringList) -> StringList:
    """Returns list of duplicates in <nameList>"""

    Logging.trace(">>: %s", nameList)

    nameSet = set()
    result = []

    for name in nameList:
        if name in nameSet:
            result.append(name)

        nameSet.add(name)

    Logging.trace("<<: %s", result)
    return result

#--------------------

def _sanitizeSampleNameForFileName (sampleName : SoundFontName,
                                    channelIsKept : Boolean) -> String:
    """Replaces all characters in <sampleName> that may not be in a
       file name; if <channelIsKept> is not set, the last L or R
       letter is removed"""

    Logging.trace(">>: sampleName = '%s', channelIsKept = %s",
                  sampleName, channelIsKept)

    result = (str(sampleName)
              .replace(" ", "_")
              .replace(":", "_")
              .replace("?", "_")
              .replace("/", "_"))

    if not channelIsKept:
        lastCharacter = result[-1].upper()

        if lastCharacter in ("L", "R"):
            result = result[:-1]

        result = result.rstrip("_")

    Logging.trace("<<: '%s'", result)
    return result

#--------------------

_writeMessage = OperatingSystem.showMessageOnConsole

#====================

class _CommandLineOptions:
    """This module handles command line options and checks them."""

    #--------------------

    @classmethod
    def checkArguments (cls,
                        argumentList : Object):
        """Checks whether command line options given in <argumentList>
           are okay"""

        Logging.trace(">>")

        loggingFilePath       = argumentList.loggingFilePath
        soundFontFilePath     = argumentList.soundFontFilePath
        waveFileDirectoryPath = argumentList.waveFileDirectoryPath

        if soundFontFilePath is not None:
            ValidityChecker.isReadableFile(soundFontFilePath,
                                           "soundFontFilePath")

        if loggingFilePath is not None:
            ValidityChecker.isWritableFile(loggingFilePath,
                                           "loggingFilePath")

        if waveFileDirectoryPath is not None:
            ValidityChecker.isDirectory(waveFileDirectoryPath,
                                        "waveFileDirectoryPath")

        Logging.trace("<<")

    #--------------------

    @classmethod
    def read (cls):
        """Reads commandline options and sets variables appropriately;
           returns tuple of variables read"""

        Logging.trace(">>")

        programDescription = ("Generates JSON output and optionally"
                              + " sample wave files from a SoundFont file")
        p = argparse.ArgumentParser(description=programDescription)

        p.add_argument("-i",
                       dest = "soundFontFilePath",
                       required = True,
                       help = ("defines the path of the SoundFont"
                               + " source file to be converted"))
        p.add_argument("-idx", "--indexing_kind",
                       dest = "objectIDIndexingKind",
                       choices = ["uuid", "nat" ],
                       default = "nat",
                       help = ("tells the kind of identification indexing"
                               + " for all objects: 'uuid' generates"
                               + " unique identification indices, 'nat'"
                               + " numbers the objects with natural"
                               + " numbers starting at zero"))
        p.add_argument("-l", "--logging_file",
                       dest = "loggingFilePath",
                       help = ("defines the path for the logging file;"
                              + " activates logging when given"))
        p.add_argument("-sf", "--sample_format",
                       dest = "sampleFormatKind",
                       choices = ["n", "i16", "i24", "i32"],
                       default = "n",
                       help = ("gives the sample format of the wave files"
                               + " (if they are generated): 'i16' is a"
                               + " 16-bit integer format, 'i24' is a 24-bit"
                               + " integer format, 'i32' a 32-bit integer"
                               + " format; 'n' stands for the native format"
                               + " (either i16 or i24 depending on the"
                               + " SoundFont version)"))
        p.add_argument("-sn", "--sample_naming",
                       dest = "namingForSampleFiles",
                       choices = ["i", "n"],
                       default = "n",
                       help = ("tells the naming conventions for the wave"
                               + " file names (if they are generated):"
                               + " 'i' uses the sample identification,"
                               + " 'n' uses the sample name"))
        p.add_argument("-w", "--wave_file_directory",
                       dest = "waveFileDirectoryPath",
                       help = ("defines the (optional) path to the"
                               + " target directory for wave files;"
                               + " if this option is not set, only the"
                               + " plain JSON output is produced"))
        result = p.parse_args()

        Logging.trace("<<: %s", result)
        return result

#====================

def _convertSoundFontToJSON (soundFontFilePath : String,
                             waveFileDirectoryPath : String,
                             namingForSampleFiles : String,
                             objectIDIndexingKind : String,
                             sampleFormatKind : String):
    """Writes information about SoundFont in <soundFontFile> to
       standard output; wave files are written to
       <waveFileDirectoryPath> (if set) using <namingForSampleFiles>
       to describe the naming convention and <sampleFormatKind> for
       the format of the sample files; <objectIDIndexingKind> tells
       the kind of identification indexing kind for all objects"""

    Logging.trace(">>: soundFontFilePath = '%s',"
                  + " waveFileDirectoryPath = '%s',"
                  + " namingForSampleFiles = '%s',"
                  + " objectIDIndexingKind = '%s',"
                  + " sampleFormatKind = '%s'",
                  soundFontFilePath, waveFileDirectoryPath,
                  namingForSampleFiles, objectIDIndexingKind,
                  sampleFormatKind)

    soundFontFileReader = \
        SoundFontFileReader(soundFontFilePath)
    soundFont = soundFontFileReader.readSoundFont(objectIDIndexingKind)
    waveFilesAreWritten = (waveFileDirectoryPath is not None)
    SoundFontSample.enableNormalizedWaveReferences(waveFilesAreWritten)
    propertyMap = soundFont.toPropertyMap()
    lineList = json.dumps(propertyMap, indent = 4).split("\n")
    sys.stdout.write(_newline.join(lineList) + _newline)

    if waveFilesAreWritten:
        # ensure that there are no duplicates in the sample names
        nameList = [ sample.name for sample in soundFont.sampleList ]
        duplicateNameList = _findDuplicateNames(nameList)

        if len(duplicateNameList) == 0:
            _writeWaveFiles(soundFont, waveFileDirectoryPath,
                            namingForSampleFiles, sampleFormatKind)
        else:
            for name in duplicateNameList:
                _writeMessage(_ErrMsg_duplicateSampleName % name)

    Logging.trace("<<")

#--------------------

def _writeSampleWaveFile (soundFont : SoundFont,
                          sampleList : ObjectList,
                          waveFilePath : String,
                          sampleFormatKind : String):
    """Writes a wave file named <waveFilePath> with wave data for all
       sample headers in <sampleList> belonging to <soundFont>
       using <sampleFormatKind> for the sample format"""

    Assertion.pre(len(sampleList) > 0,
                  "sample header list must not be empty")
    
    Logging.trace(">>: sampleList = %s, waveFilePath = '%s',"
                  + " sampleFormatKind = '%s'",
                  sampleList, waveFilePath, sampleFormatKind)

    # sample settings
    waveData = soundFont.waveData
    widthSpec = sampleFormatKind[1:]

    sampleKind = "i"
    sampleWidth = (2      if widthSpec == "16"
                   else 3 if widthSpec == "24"
                   else 4 if widthSpec == "32"
                   else waveData.bytesPerDataPoint())

    getSampleWaveData = (lambda sample:
                         waveData.slice(sample.sampleStartPosition,
                                        sample.sampleEndPosition))

    # parameters for the wave file
    firstSample  = sampleList[0]
    channelCount = len(sampleList)
    sampleBuffer = [ getSampleWaveData(sample) for sample in sampleList ]
    sampleCount  = len(sampleBuffer[0])
    sampleRate   = firstSample.sampleRate

    waveFile = WaveFile(waveFilePath)
    waveFile.write(channelCount, sampleCount, sampleBuffer,
                   sampleRate, sampleWidth, sampleKind)

    Logging.trace("<<")

#--------------------

def _writeWaveFiles (soundFont : SoundFont,
                     waveFileDirectoryPath : String,
                     namingForSampleFiles : String,
                     sampleFormatKind : String):
    """Writes wave files in <soundFont> to <waveFileDirectoryPath>
       using <namingForSampleFiles> to describe the naming convention
       and <sampleFormatKind> for the sample format"""

    Logging.trace(">>: waveFileDirectoryPath = '%s',"
                  + " namingForSampleFiles = '%s',"
                  + " sampleFormatKind = '%s'",
                  waveFileDirectoryPath, namingForSampleFiles,
                  sampleFormatKind)

    sampleList = soundFont.sampleList
    getSampleCount = \
        lambda sample: sample.sampleEndPosition - sample.sampleStartPosition
    sampleIdSet = set()
   
    for sample in sampleList:
        identification = sample.identification
        name           = sample.name

        if identification not in sampleIdSet:
            Logging.trace("--: processing sample '%s' - '%s'",
                          name, identification)
            sampleIdSet.add(identification)

            fileName = (identification if namingForSampleFiles == "i"
                        else name)
            fileName = _sanitizeSampleNameForFileName(fileName, True)
            partnerSample = sample.partner
            hasPartner = partnerSample is not None
            sampleList = []

            if not hasPartner:
                sampleList.append(sample)
            else:
                sampleKind = sample.kind
                sampleCount = getSampleCount(sample)
                otherSampleCount = getSampleCount(partnerSample)

                if sampleCount != otherSampleCount:
                    Logging.trace("--: linked samples have different"
                                  + " lengths => split them")
                    sampleList.append(sample)
                else:
                    fileName = _sanitizeSampleNameForFileName(name, False)
                    sampleIdSet.add(partnerSample.identification)

                    if sampleKind == SoundFontSampleKind.leftSample:
                        sampleList.append(sample)
                        sampleList.append(partnerSample)
                    else:
                        sampleList.append(partnerSample)
                        sampleList.append(sample)

            Logging.trace("--: identification = '%s', channelCount = %d",
                          identification, len(sampleList))

            # write data to file
            waveFilePath = "%s/%s.wav" % (waveFileDirectoryPath, fileName)
            _writeSampleWaveFile(soundFont, sampleList, waveFilePath,
                                 sampleFormatKind)

    Logging.trace("<<")

#====================

def _finalize ():
    Logging.finalize()

#--------------------

def _initialize ():
    Logging.initialize()
    Logging.setTracingWithTime(True, 3)

#--------------------
#--------------------

def _process (soundFontFilePath : String,
              loggingFilePath : String,
              waveFileDirectoryPath : String,
              namingForSampleFiles : String,
              objectIDIndexingKind : String,
              sampleFormatKind : String):
    """Processes SoundFont file named <soundFontFilePath>, does
       logging to <loggingFilePath> (if set) and writes wave files to
       <waveFileDirectoryPath> (if set) using <namingForSampleFiles>
       to describe the naming convention; <objectIDIndexingKind> tells
       the kind of identification indexing kind for all objects"""

    if loggingFilePath is None:
        Logging.setLevel(Logging_Level.noLogging)
    else:
        Logging.setFileName(loggingFilePath)
        Logging.setLevel(Logging_Level.verbose)

    Logging.trace(">>: soundFontFilePath = '%s',"
                  + " loggingFilePath = '%s',"
                  + " waveFileDirectoryPath = '%s',"
                  + " namingForSampleFiles = '%s',"
                  + " objectIDIndexingKind = '%s',"
                  + " sampleFormatKind = '%s'",
                  soundFontFilePath, loggingFilePath,
                  waveFileDirectoryPath, namingForSampleFiles,
                  objectIDIndexingKind, sampleFormatKind)

    _convertSoundFontToJSON(soundFontFilePath, waveFileDirectoryPath,
                            namingForSampleFiles, objectIDIndexingKind,
                            sampleFormatKind)

    Logging.trace("<<")

#--------------------
#--------------------

def main ():
    """The main program"""

    _initialize()
    argumentList = _CommandLineOptions.read()
    _CommandLineOptions.checkArguments(argumentList)

    _process(argumentList.soundFontFilePath,
             argumentList.loggingFilePath,
             argumentList.waveFileDirectoryPath,
             argumentList.namingForSampleFiles,
             argumentList.objectIDIndexingKind,
             argumentList.sampleFormatKind)

    _finalize()

#--------------------
#--------------------

if __name__ == "__main__":
    main()



# soundFontAnalyser - scans SoundFont file for possible problems
#
# author: Dr. Thomas Tensi
# version: 2025-08

#====================
# IMPORTS
#====================

import argparse
from enum import StrEnum
import json
import math
import sys

from basemodules.configurationfile import ConfigurationFile
from basemodules.operatingsystem import OperatingSystem
from basemodules.simplelogging import Logging, Logging_Level
from basemodules.simpletypes import \
    Boolean, Class, Natural, Object, ObjectList, Real, String, \
    StringList, Tuple
from basemodules.stringutil import deserializeToList
from basemodules.ttbase import iif, iif3
from basemodules.validitychecker import ValidityChecker

from simplefunctions import \
    OneDimensionalFunctionFromIntervals, \
    OneDimensionalFunctionFromNumbers, \
    TwoDimensionalFunctionFromIntervals

from multimedia.midi.soundfontfile import SoundFontFileReader

from multimedia.midi.soundfont import \
    SoundFont, SoundFontGeneratorAmount, \
    SoundFontGeneratorKind, SoundFontObjectKind, \
    _SoundFontIdentifiedElement, SoundFontZonedElement

#====================
# Abbreviations
#====================

ODFFI = OneDimensionalFunctionFromIntervals
SFGA = SoundFontGeneratorAmount
SFGK = SoundFontGeneratorKind

#====================
# TYPES
#====================

GeneratorKindList         = ObjectList
SoundFontZonedElementList = ObjectList
SoundFontZoneList         = ObjectList
Pair                      = Tuple

#====================

_programName = "soundFontAnalyser"
_newline = "\n"

_partnerGeneratorKindList = (SoundFontGeneratorKind.sampleID,
                             SoundFontGeneratorKind.instrument)

#--------------------

# the name of all rule checks
_ruleNameList = (
    "instrumentGlobalZone", "instrumentModulators",
    "instrumentOverrides",
    "presetGlobalZone", "presetModulators",
    "sampleOverrides"
)

#--------------------

_modulatorGeneratorKindList = [
    SFGK.startAddrsOffset, SFGK.endAddrsOffset, SFGK.startLoopAddrsOffset,
    SFGK.endLoopAddrsOffset, SFGK.startAddrsCoarseOffset,
    SFGK.modLfoToPitch, SFGK.vibLfoToPitch, SFGK.modEnvToPitch,
    SFGK.initialFilterFc, SFGK.initialFilterQ, SFGK.modLfoToFilterFc,
    SFGK.modEnvToFilterFc, SFGK.endAddrsCoarseOffset, SFGK.modLfoToVolume,
    SFGK.chorusEffectsSend, SFGK.reverbEffectsSend, SFGK.pan,
    SFGK.delayModLFO, SFGK.freqModLFO, SFGK.delayVibLFO, SFGK.freqVibLFO,
    SFGK.delayModEnv, SFGK.attackModEnv, SFGK.holdModEnv, SFGK.decayModEnv,
    SFGK.sustainModEnv, SFGK.releaseModEnv,
    SFGK.keynumToModEnvHold, SFGK.keynumToModEnvDecay,
    SFGK.delayVolEnv, SFGK.attackVolEnv, SFGK.holdVolEnv,
    SFGK.decayVolEnv, SFGK.sustainVolEnv, SFGK.releaseVolEnv,
    SFGK.keynumToVolEnvHold, SFGK.keynumToVolEnvDecay,
    SFGK.keynum, SFGK.velocity, SFGK.initialAttenuation,
    SFGK.coarseTune, SFGK.fineTune, SFGK.sampleModes,
    SFGK.scaleTuning, SFGK.exclusiveClass, SFGK.overridingRootKey
]

#--------------------

# all generator kinds
_allGeneratorKindList = list(SoundFontGeneratorKind)

#--------------------

# a mapping from rule name to generator kind list
_ruleNameToGeneratorKindListMap = {
    "instrumentGlobalZone" : _allGeneratorKindList,

    "instrumentModulators" : _modulatorGeneratorKindList,

    "instrumentOverrides" : [
        SFGK.keyRange, SFGK.velRange,
        SFGK.coarseTune, SFGK.fineTune, SFGK.scaleTuning,
        SFGK.chorusEffectsSend, SFGK.reverbEffectsSend
    ],

    "presetGlobalZone" : _allGeneratorKindList,

    "presetModulators" : _modulatorGeneratorKindList,
    
    "sampleOverrides" : [
        SFGK.overridingRootKey, SFGK.startAddrsOffset,
        SFGK.startLoopAddrsOffset
    ]
}

#====================

class _CommandLineOptions:
    """This module handles command line options and checks them."""

    #--------------------

    @classmethod
    def checkArguments (cls,
                        argumentList : Object):
        """Checks whether command line options given in <argumentList>
           are okay"""

        Logging.trace(">>")

        ValidityChecker.isReadableFile(argumentList.soundFontFilePath,
                                       "soundFontFilePath")

        loggingFilePath       = argumentList.loggingFilePath
        configurationFilePath = argumentList.configurationFilePath

        if loggingFilePath is not None:
            ValidityChecker.isWritableFile(loggingFilePath,
                                           "loggingFilePath")

        if configurationFilePath is not None:
            ValidityChecker.isReadableFile(configurationFilePath,
                                           "configurationFilePath")

        Logging.trace("<<")

    #--------------------

    @classmethod
    def read (cls):
        """Reads commandline options and sets variables appropriately;
           returns tuple of variables read"""

        Logging.trace(">>")

        programDescription = \
            "Analyses SoundFont file and writes results to standard output"
        p = argparse.ArgumentParser(description=programDescription)

        p.add_argument("-c",
                       dest = "configurationFilePath",
                       help = ("defines the path of the (optional)"
                               + " configuration file"))
        p.add_argument("-l", "--logging_file",
                       dest = "loggingFilePath",
                       help = ("defines the path for the logging file;"
                              + " activates logging when given"))
        p.add_argument("-p", action = "store_true",
                       dest = "outputIsPrettyPrinted",
                       help = ("tells to format output in a" +
                               " pretty-printed style (human-readable)"))
        p.add_argument("soundFontFilePath",
                       help = ("defines the path of the SoundFont"
                               + " source file to be analysed"))

        result = p.parse_args()

        Logging.trace("<<: %s", result)
        return result

#====================

class _ConfigurationManager:
    """This module encapsulates the data from the configuration file."""

    # the set of excluded rule names (from configuration file)
    _excludedRuleNameSet = set()

    # mapping from rule name to excluded generator kind set
    _ruleNameToExcludedGeneratorKindSetMap = {}
    
    # setting for maximum accepted function distance when modulator
    # shall be recommended
    _maximumNormalizedRootMeanSquareError = 0.3

    # setting for maximum relative difference between two values when
    # projecting a two-dimensional generator function to one dimension
    _comparisonMarginForOneDFunctions = 0.2

    #--------------------
    #--------------------

    @classmethod
    def adaptGeneratorKindList (cls : Class,
                                ruleName : String,
                                generatorKindList : GeneratorKindList):
        """Strips generator kinds excluded in configuration file for
           rule named <ruleName> from <generatorKindList>"""

        Logging.trace(">>: ruleName = %s, generatorKindList = %s",
                      ruleName, generatorKindList)

        nameToExcludedGenKindSetMap = \
            cls._ruleNameToExcludedGeneratorKindSetMap

        if ruleName in nameToExcludedGenKindSetMap:
            excludedGeneratorKindSet = \
                nameToExcludedGenKindSetMap[ruleName]

            tempList = []

            for generatorKind in generatorKindList:
                if generatorKind not in excludedGeneratorKindSet:
                    tempList.append(generatorKind)

            generatorKindList.clear()
            generatorKindList.extend(tempList)

        Logging.trace("<<: %s", generatorKindList)

    #--------------------

    @classmethod
    def read (cls : Class,
              configurationFilePath : String):
        """Sets up manager with configuration file path
           <configurationFilePath>"""

        Logging.trace(">>: '%s'", configurationFilePath)

        configurationFile = ConfigurationFile(configurationFilePath)

        # process real bound variables
        def readRealProc (key : String, defaultValue : Real) -> Tuple:
            st = configurationFile.value(key, defaultValue)
            isOkay = ValidityChecker.isReal(st, key, False)

            if not isOkay:
                Logging.traceError("bad value %s for '%s' in configuration"
                                   + " file",
                                   st, key)

            return isOkay, st

        #-----

        isOkay, st = readRealProc("maximumNormalizedRootMeanSquareError", 0.1)

        if isOkay:
            cls._maximumNormalizedRootMeanSquareError = float(st)

        isOkay, st = readRealProc("comparisonMarginForOneDFunctions", 0.2)

        if isOkay:
            cls._comparisonMarginForOneDFunctions = float(st)

        # process rule variables
        nameToExcludedGenKindSetMap = \
            cls._ruleNameToExcludedGeneratorKindSetMap

        for ruleName in _ruleNameList:
            keyPrefix = "rule_%s_" % ruleName

            # check whether rule is active
            key = keyPrefix + "isActive"
            isActive = configurationFile.value(key, True)
            
            if not isActive:
                cls._excludedRuleNameSet.add(ruleName)

            # find list of excluded generator kinds
            key = keyPrefix + "excludedGeneratorKindList"
            kindListAsString = configurationFile.value(key, "[]")
            stringList = deserializeToList(kindListAsString)
            generatorKindSet = set()

            for st in stringList:
                generatorKind = SoundFontGeneratorKind.fromString(st)

                if generatorKind is not None:
                    generatorKindSet.add(generatorKind)

            nameToExcludedGenKindSetMap[ruleName] = generatorKindSet

        Logging.trace("--: _excludedRuleNameSet = %s,"
                      + " _ruleNameToExcludedGeneratorKindSetMap = %s",
                      cls._excludedRuleNameSet,
                      cls._ruleNameToExcludedGeneratorKindSetMap)

        Logging.trace("<<")

    #--------------------

    @classmethod
    def ruleIsChecked (cls : Class,
                       ruleName : String) -> Boolean:
        """Tells whether rule named <ruleName> is active in
           configuration file"""

        Logging.trace(">>: ruleName = %s", ruleName)
        result = ruleName not in cls._excludedRuleNameSet
        Logging.trace("<<: %s", result)
        return result

#====================

class _SoundFontAnalyser:
    """Provides the checking of all rules for a SoundFont analysis"""

    #--------------------
    # PRIVATE FEATURES
    #--------------------

    def _addMessage (self : Object,
                     objectKind : SoundFontObjectKind,
                     objectName : String,
                     message : String):
        """Adds analysis message <message> for SoundFont object with
           <objectKind> and name <objectName> to internal data"""

        Logging.trace(">>: objectKind = '%s', objectName = '%s',"
                      + " message = '%s'",
                      objectKind, objectName, message)

        key = (objectKind, objectName)
        nameToMessageListMap = self._nameToMessageListMap

        if key not in nameToMessageListMap:
            nameToMessageListMap[key] = []

        nameToMessageListMap[key].append(message)

        Logging.trace("<<")
    
    #--------------------

    def _analyse1DFunctionForModulator \
            (self : Object,
             objectKind : SoundFontObjectKind,
             objectName : String,
             controllingParameter : String,
             generatorKind : SoundFontGeneratorKind,
             intervalFunction : OneDimensionalFunctionFromIntervals):
        """Checks whether <intervalFunction> for generator of kind
           <generatorKind> from interval to real for <objectKind> and
           <objectName> can be implemented by a modulator controlled
           by <controllingParameter>"""

        generatorKindString = generatorKind.toShortString()
        Logging.trace(">>: objectKind = %s, objectName = '%s',"
                      " generatorKind = %s, controllingParameter = '%s'"
                      + " function = %s",
                      objectKind, objectName, generatorKindString,
                      controllingParameter, intervalFunction)

        # the modulation factor is always relative, hence we assume
        # that we are in a preset zone
        valueAdaptationProc = \
            SFGA.wordFromRealProc(generatorKind, False)
        approximationResult = \
            ODFFI.approximate(intervalFunction, valueAdaptationProc)
        curveKind, functionDistance, factor, adaptedFactor, offset, \
            isUnipolar, isAbsolute, isAscending = \
                approximationResult.asTuple()
        evaluationMessage = None
        maximumFunctionDistance = \
            _ConfigurationManager._maximumNormalizedRootMeanSquareError

        if curveKind is None:
            evaluationMessage = ("%s: no approximation found for %s"
                                 % (objectName, generatorKindString))
        elif functionDistance > maximumFunctionDistance:
            evaluationMessage = \
                (("%s: skipped approximation for %s because"
                  + " function distance %.3f is greater than %.3f")
                 % (objectName, generatorKindString,
                    functionDistance, maximumFunctionDistance))
        else:
                messageTemplate = \
                    ("data for generator kind '%s' can be provided by"
                     + " a global modulator"
                     + " (controllerKind = '%s',"
                     + " isUnipolar = %s, isAscending = %s, curveKind = %s,"
                     + " factor = %.3f (=> %s), isAbsolute = %s)"
                     + " and offset = %.3f; with an overall function"
                     + " distance of %.4f, new function = %s")
                newFunction = \
                    ODFFI.make(intervalFunction.domainValueList(),
                               factor, offset,
                               isUnipolar, isAscending, curveKind, isAbsolute)
                newFunctionAsString = \
                    self._makeComparativeString(newFunction, intervalFunction)

                message = (messageTemplate
                           % (generatorKindString,
                              controllingParameter,
                              isUnipolar, isAscending, curveKind,
                              factor, adaptedFactor, isAbsolute,
                              offset,
                              functionDistance, newFunctionAsString))
                self._addMessage(objectKind, objectName, message)
                evaluationMessage = message

        if evaluationMessage is not None:
            Logging.trace("--: EVALUATION - %s", evaluationMessage)

        Logging.trace("<<")

    #--------------------

    def _analyse2DFunctionForModulator \
            (self : Object,
             objectKind : SoundFontObjectKind,
             objectName : String,
             generatorKind : SoundFontGeneratorKind,
             function : TwoDimensionalFunctionFromIntervals):
        """Checks whether <function> from key range and velocity range
           to value can be implemented by a modulator"""

        generatorKindString = generatorKind.toShortString()
        Logging.trace(">>: objectName = '%s', generatorKind = %s,"
                      + " function = %s",
                      objectName, generatorKindString, function)

        relevantFunction = None
        comparisonMarginForOneDFunctions = \
            int(100 * _ConfigurationManager._comparisonMarginForOneDFunctions)

        if function.isOneDimensional(comparisonMarginForOneDFunctions):
            relevantFunction = function
            parameter = "velocity"
        else:
            otherFunction = function.flip()

            if otherFunction.isOneDimensional(comparisonMarginForOneDFunctions):
                relevantFunction = otherFunction
                parameter = "key"

        if relevantFunction is not None:
            firstInterval = relevantFunction.domainValueList()[0]
            oneDimensionalFunction = relevantFunction.at(firstInterval)
            Logging.trace("--: function ('%s', %s)  only depends"
                          + " on %s: %s",
                          objectName, generatorKindString, parameter,
                          oneDimensionalFunction)

            self._analyse1DFunctionForModulator(objectKind,
                                                objectName,
                                                parameter,
                                                generatorKind,
                                                oneDimensionalFunction)

        Logging.trace("<<")

    #--------------------

    def _checkForPossibleGlobalZoneOverrides \
                        (self : Object,
                         objectKind : SoundFontObjectKind,
                         zonedObjectList : SoundFontZonedElementList):
        """Checks whether there are identical settings in the
           non-global zones of some object in <zonedObjectList> that
           can be generalized to the global zone"""

        Logging.trace(">>: objectKind = '%s', listCount = %d",
                      objectKind, len(zonedObjectList))

        messageTemplate = "data for '%s' can be moved to global zone"
        ruleName = "%sGlobalZone" % objectKind
        valueGeneratorKindList = \
            _ruleNameToGeneratorKindListMap[ruleName]
        _ConfigurationManager \
            .adaptGeneratorKindList(ruleName, valueGeneratorKindList)

        for zonedObject in zonedObjectList:
            objectName = zonedObject.name
            Logging.trace("--: analyzing %s '%s'",
                          objectKind, objectName)

            if len(zonedObject.zoneList) < 2:
                Logging.trace("--: skipped, because there are less"
                              + " than two non-global zones")
                continue

            for generatorKind in valueGeneratorKindList:
                if generatorKind in _partnerGeneratorKindList:
                    # those kinds identify a non-global zone and
                    # must not be consolidated
                    continue

                valueList = \
                    self._collectZoneValuesAsList(zonedObject,
                                                  generatorKind)

                if len(valueList) == 0:
                    continue

                firstValue = valueList[0]
                hasIdenticalValues = \
                    all([ value == firstValue for value in valueList ])

                Logging.trace("--: valueList = %s, isEqual = %s",
                              valueList, hasIdenticalValues)

                if hasIdenticalValues:
                    message = \
                        messageTemplate % generatorKind.toShortString()
                    self._addMessage(objectKind, objectName, message)

        Logging.trace("<<")

    #--------------------

    def _checkForPossibleZoneModulations \
            (self : Object,
             objectKind : SoundFontObjectKind,
             zonedObjectList : SoundFontZonedElementList):
        """Checks whether there are settings in the non-global zones
           of some object in <zonedObjectList> that can be generalized
           via a modulator"""

        Logging.trace(">>: objectKind = '%s', listCount = %d",
                      objectKind, len(zonedObjectList))

        ODFFI.initialize()
        ruleName = "%sModulators" % objectKind
        valueGeneratorKindList = \
            _ruleNameToGeneratorKindListMap[ruleName]
        _ConfigurationManager \
            .adaptGeneratorKindList(ruleName, valueGeneratorKindList)

        for zonedObject in zonedObjectList:
            objectName = zonedObject.name
            Logging.trace("--: analyzing %s '%s'", objectKind, objectName)

            if len(zonedObject.zoneList) < 4:
                Logging.trace("--: skipped, because there are less than"
                              + " four non-global zones")
                continue

            for generatorKind in valueGeneratorKindList:
                function = \
                    self._collectZoneValuesAs2DFunction(zonedObject,
                                                        generatorKind)

                if (function.isEmpty()
                    or function.isConstant()
                    or not function.isConsistent()):
                    Logging.trace("--: function is empty, constant"
                                  + " or inconsistent")
                else:
                    self._analyse2DFunctionForModulator(objectKind,
                                                        objectName,
                                                        generatorKind,
                                                        function)

        Logging.trace("<<")

    #--------------------

    def _collectZoneValuesAsList \
            (self : Object,
             zonedObject : SoundFontZonedElement,
             valueGeneratorKind : SoundFontGeneratorKind) -> ObjectList:
        """Collects values for <generatorKind> from non-global zones
           for <zonedObject> into object list and returns it"""

        Logging.trace(">>: zonedObject = '%s', valueGeneratorKind = '%s'",
                      zonedObject.name, valueGeneratorKind)

        globalZone = zonedObject.globalZone
        zoneList   = zonedObject.zoneList

        getValue = lambda x, kind: x.generatorMap.get(kind)
        globalValue = (None if globalZone is None
                       else getValue(globalZone, valueGeneratorKind))
        valueList = [ getValue(zone, valueGeneratorKind)
                      for zone in zoneList ]
        hasValues = any([ value is not None for value in valueList])

        if not hasValues:
            Logging.trace("--: no values in zones")
            result = []
        else:
            Logging.trace("--: original values = %s", valueList)
            update = lambda x: x if x is not None else globalValue
            result = list(map(update, valueList))

        Logging.trace("<<: %s", result)
        return result

    #--------------------

    def _collectZoneValuesAs2DFunction \
            (self : Object,
             zonedObject : SoundFontZonedElement,
             valueGeneratorKind : SoundFontGeneratorKind) \
             -> TwoDimensionalFunctionFromIntervals:
        """Collects values for <generatorKind> from non-global zones
           for <zonedObject> into a 2D function object and returns it"""

        Logging.trace(">>: zonedObject = '%s', valueGeneratorKind = '%s'",
                      zonedObject.name, valueGeneratorKind)

        getValue   = lambda x, kind: x.generatorMap.get(kind)
        adaptValue = (lambda x, default:
                      default if x is None else x)

        globalZone = zonedObject.globalZone
        zoneList   = zonedObject.zoneList

        keyRangeKind      = SoundFontGeneratorKind.keyRange
        velocityRangeKind = SoundFontGeneratorKind.velRange
        fullRange = (0, 127)

        globalKeyRange = \
            adaptValue((None if globalZone is None
                        else getValue(globalZone, keyRangeKind)),
                       fullRange)
        globalVelocityRange = \
            adaptValue((None if globalZone is None
                        else getValue(globalZone, velocityRangeKind)),
                       fullRange)
        globalValue = \
            adaptValue(getValue(globalZone, valueGeneratorKind), None)

        result = TwoDimensionalFunctionFromIntervals()

        for zone in zoneList:
            keyRange = \
                adaptValue(getValue(zone, keyRangeKind), globalKeyRange)
            velocityRange = \
                adaptValue(getValue(zone, velocityRangeKind),
                           globalVelocityRange)
            value = \
                adaptValue(getValue(zone, valueGeneratorKind), globalValue)

            if value is not None:
                result.setAt(keyRange, velocityRange, value)
            else:
                result.clear()
                break

        result.consolidate()

        Logging.trace("<<: %s", result)
        return result

    #--------------------

    def _combinedZoneList (self : Object,
                           isPresetList : Boolean) -> SoundFontZoneList:
        """Returns list of all zones either for presets or instruments
           (depending on <isPresetList>"""

        Logging.trace(">>: isPresetList = %s", isPresetList)

        zonedElementList = iif(isPresetList,
                               self._soundFont.presetList,
                               self._soundFont.instrumentList)
        result = []

        for zonedElement in zonedElementList:
            result.extend(zonedElement.zoneList)

        Logging.trace("<<: count = %d", len(result))
        return result

    #--------------------

    def _makeComparativeString \
            (self : Object,
             functionA : OneDimensionalFunctionFromIntervals,
             functionB : OneDimensionalFunctionFromIntervals) -> String:
        """Returns a string with values of <functionA> and <functionB>
           combined for a comparison; assumes that the domain of both
           functions is identical"""

        Logging.trace(">>")

        result = ""

        for domainValue in functionA.domainValueList():
            result += ("%s%d..%d : %.3f (for %.3f)"
                       % (iif(result == "", "", ", "),
                          domainValue[0], domainValue[1],
                          functionA.at(domainValue),
                          functionB.at(domainValue)))

        result = "{ %s }" % result
       
        Logging.trace("<<: %s", result)
        return result

    #--------------------

    def _scanForRedundantValues (self : Object,
                                 objectKind : SoundFontObjectKind,
                                 zoneList : SoundFontZoneList,
                                 partnerGeneratorKind : SoundFontGeneratorKind,
                                 valueGeneratorKind : SoundFontGeneratorKind):
        """Analyses zones in <zoneList> for identical values for
           <valueGeneratorKind> referencing underlying elements
           characterized by <partnerGeneratorKind> and reports any
           redundancies"""

        Logging.trace(">>: objectKind = '%s', zoneCount = %d,"
                      + " partnerGeneratorKind = %s,"
                      + " valueGeneratorKind = %s",
                      objectKind, len(zoneList),
                      partnerGeneratorKind, valueGeneratorKind)

        # collect value set for value generator kind grouped by
        # partner
        identificationToValueSetMap = {}
        identificationToUsageCountMap = {}
        idSeparator = " "
        undefinedValue = -1E10
        defaultValue = valueGeneratorKind.defaultValue()

        for zone in zoneList:
            associatedGlobalZone = zone.parent.globalZone
            generatorMap = zone.generatorMap
            value = generatorMap.get(valueGeneratorKind, defaultValue)

            if value is not None:
                partner = generatorMap.get(partnerGeneratorKind)

                if partner is None:
                    continue

                identification = partner.identification

                if identification in identificationToValueSetMap:
                    identificationToUsageCountMap[identification] += 1
                else:
                    identificationToValueSetMap[identification] = set()
                    identificationToUsageCountMap[identification] = 1

                valueSet = identificationToValueSetMap[identification]

                if (associatedGlobalZone is not None
                    and associatedGlobalZone.generatorMap is not None):
                    globalGeneratorMap = associatedGlobalZone.generatorMap
                    globalValue = globalGeneratorMap.get(valueGeneratorKind)

                    if globalValue is not None and globalValue != value:
                        # value overrides global value, hence generate
                        # some dummy value for the value set
                        Logging.trace("--: override of global value")
                        value = undefinedValue
                
                valueSet.add(value)

        Logging.trace("--: identificationToValueSetMap = %s",
                      identificationToValueSetMap)

        # check for single valued sets that may be transferred to
        # underlying partner
        messageTemplate = ("all %d usage(s) have identical values"
                           + " for '%s' - %r")
        messageCount = 0

        for identification, valueSet in identificationToValueSetMap.items():
            if len(valueSet) == 1:
                value = list(valueSet)[0]

                if value not in (undefinedValue, defaultValue):
                    usageCount = identificationToUsageCountMap[identification]
                    partner = (_SoundFontIdentifiedElement
                               .getByIdentification(identification))
                    message = (messageTemplate
                            % (usageCount,
                               valueGeneratorKind.toShortString(), value))
                    self._addMessage(partner.elementKind, partner.name,
                                     message)
                    messageCount += 1

        Logging.trace("<<: number of messages = %d", messageCount)

    #--------------------
    # EXPORTED FEATURES
    #--------------------

    def __init__ (self : Object,
                  soundFont : SoundFont):
        """Initializes analyser <self> for <soundFont>"""

        Logging.trace(">>")

        self._soundFont = soundFont
        self._nameToMessageListMap = {}

        Logging.trace("<<")

    #--------------------
    # access
    #--------------------

    def getResults (self : Object,
                    outputIsPrettyPrinted : Boolean) -> StringList:
        """Returns line list with analysis results"""

        Logging.trace(">>: outputIsPrettyPrinted = %s",
                      outputIsPrettyPrinted)

        result = []

        kindToPosition = lambda x: iif3(x == "header", 0,
                                        x == "sample", 1,
                                        x == "instrument", 2, 3)
        keyToPosition = lambda x: (kindToPosition(x[0]), str(x[1]))
        keyList = sorted(self._nameToMessageListMap.keys(),
                         key = keyToPosition)

        previousObjectKind = ""
        indentation = " " * 4

        for key in keyList:
            objectKind, objectName = key
            messageList = self._nameToMessageListMap[key]

            if outputIsPrettyPrinted:
                if previousObjectKind != objectKind:
                    previousObjectKind = objectKind

                    line = ("=== %s%s ==="
                            % (objectKind.upper(),
                               "S" if objectKind != "header" else ""))
                    result.append(line)

                result.append("%s%s" % (indentation, objectName))

            for message in messageList:
                if not outputIsPrettyPrinted:
                    line = "%s '%s': %s" % (objectKind, objectName, message)
                else:
                    line = indentation * 2 + message

                result.append(line)
                    
        Logging.trace("<<")
        return result
    
    #--------------------
    # rule checkers
    #--------------------

    def checkInstrumentGlobalZone (self : Object):
        """Checks whether there are identical settings in the voice
           zones of an instrument that can be generalized to the
           global zone"""

        Logging.trace(">>")

        ruleName = "instrumentGlobalZone"

        if _ConfigurationManager.ruleIsChecked(ruleName):
            instrumentList = self._soundFont.instrumentList
            self._checkForPossibleGlobalZoneOverrides \
                (SoundFontObjectKind.instrument, instrumentList)

        Logging.trace("<<")

    #--------------------

    def checkInstrumentModulations (self : Object):
        """Checks whether there are settings in the voice zones of an
           instrument that can be generalized via a modulator"""

        Logging.trace(">>")

        ruleName = "instrumentModulators"

        if _ConfigurationManager.ruleIsChecked(ruleName):
            instrumentList = self._soundFont.instrumentList
            self._checkForPossibleZoneModulations \
                (SoundFontObjectKind.instrument, instrumentList)

        Logging.trace("<<")

    #--------------------

    def checkInstrumentOverrideParameters (self : Object):
        """Checks whether there are identical settings in the voice
           zones of a preset that can be applied to the
           underlying instrument"""

        Logging.trace(">>")

        ruleName = "instrumentOverrides"

        if _ConfigurationManager.ruleIsChecked(ruleName):
            presetZoneList = self._combinedZoneList(True)
            partnerGeneratorKind = SFGK.instrument
            valueGeneratorKindList = \
                _ruleNameToGeneratorKindListMap[ruleName]
            _ConfigurationManager \
                .adaptGeneratorKindList(ruleName, valueGeneratorKindList)

            for valueGeneratorKind in valueGeneratorKindList:
                self._scanForRedundantValues(SoundFontObjectKind.instrument,
                                             presetZoneList,
                                             partnerGeneratorKind,
                                             valueGeneratorKind)

        Logging.trace("<<")

    #--------------------

    def checkPresetGlobalZone (self : Object):
        """Checks whether there are identical settings in the voice
           zones of a preset that can be generalized to the global
           zone"""

        Logging.trace(">>")

        ruleName = "presetGlobalZone"
        
        if _ConfigurationManager.ruleIsChecked(ruleName):
            presetList = self._soundFont.presetList
            objectKind = SoundFontObjectKind.preset
            self._checkForPossibleGlobalZoneOverrides(objectKind,
                                                      presetList)

        Logging.trace("<<")

    #--------------------

    def checkPresetModulations (self : Object):
        """Checks whether there are settings in the voice zones of a
           preset that can be generalized via a modulator"""

        Logging.trace(">>")

        ruleName = "presetModulators"

        if _ConfigurationManager.ruleIsChecked(ruleName):
            presetList = self._soundFont.presetList
            self._checkForPossibleZoneModulations \
                (SoundFontObjectKind.preset, presetList)

        Logging.trace("<<")

    #--------------------

    def checkSampleOverrideParameters (self : Object):
        """Checks whether there are identical settings in the voice
           zones of an instrument that can be applied to the
           underlying sample"""

        Logging.trace(">>")

        ruleName = "sampleOverrides"
        
        if _ConfigurationManager.ruleIsChecked(ruleName):
            SFGK = SoundFontGeneratorKind
            instrumentZoneList = self._combinedZoneList(False)
            partnerGeneratorKind = SFGK.sampleID
            valueGeneratorKindList = \
                _ruleNameToGeneratorKindListMap[ruleName]

            _ConfigurationManager \
                .adaptGeneratorKindList(ruleName, valueGeneratorKindList)

            for valueGeneratorKind in valueGeneratorKindList:
                self._scanForRedundantValues("sample",
                                             instrumentZoneList,
                                             partnerGeneratorKind,
                                             valueGeneratorKind)

        Logging.trace("<<")

#====================

def _analyseSoundFont (soundFont : SoundFont,
                       outputIsPrettyPrinted : Boolean) -> StringList:
    """Collects analysis information about <soundFont> in a line list
       and returns it"""

    Logging.trace(">>: outputIsPrettyPrinted = %s",
                  outputIsPrettyPrinted)

    soundFontAnalyser = _SoundFontAnalyser(soundFont)

    soundFontAnalyser.checkSampleOverrideParameters()
    soundFontAnalyser.checkInstrumentGlobalZone()
    soundFontAnalyser.checkInstrumentModulations()
    soundFontAnalyser.checkInstrumentOverrideParameters()
    soundFontAnalyser.checkPresetGlobalZone()

    result = soundFontAnalyser.getResults(outputIsPrettyPrinted)
    
    Logging.trace("<<: lineCount = %d", len(result))
    return result

#--------------------

def _process (soundFontFilePath : String,
              configurationFilePath : String,
              loggingFilePath : String,
              outputIsPrettyPrinted : Boolean):
    """Processes SoundFont file named <soundFontFilePath> and writes
       analysis results to standard output; <loggingFilePath> gives
       the path to the optional logging file, <outputIsPrettyPrinted>
       tells to output in human readable form and
       <configurationFilePath> gives the path of an optional
       configuration file"""

    global _ruleNameToGeneratorKindListMap

    if loggingFilePath is None:
        Logging.setLevel(Logging_Level.noLogging)
    else:
        Logging.setFileName(loggingFilePath)
        Logging.setLevel(Logging_Level.verbose)

    Logging.trace(">>: soundFontFilePath = '%s',"
                  + " configurationFilePath = '%s',"
                  + " loggingFilePath = '%s',"
                  + " outputIsPrettyPrinted = %s",
                  soundFontFilePath, configurationFilePath,
                  loggingFilePath, outputIsPrettyPrinted)

    # read soundfont
    Logging.setLevel(Logging_Level.noLogging)
    soundFontFileReader = SoundFontFileReader(soundFontFilePath)
    soundFont = soundFontFileReader.readSoundFont()
    Logging.setLevel(Logging_Level.verbose)

    # read configuration file (if any)
    if configurationFilePath is not None:
        _ConfigurationManager.read(configurationFilePath)

    # bring all lists in Polyphone order
    for _, value in _ruleNameToGeneratorKindListMap.items():
        SFGK.adaptToPolyphoneOrder(value)
        
    # analyse soundfont
    lineList = _analyseSoundFont(soundFont, outputIsPrettyPrinted)

    # write lines to stdout
    sys.stdout.write(_newline.join(lineList) + _newline)

    Logging.trace("<<")

#====================

def _finalize ():
    Logging.finalize()

#--------------------

def _initialize ():
    Logging.initialize()
    Logging.setTracingWithTime(True, 3)

#--------------------
#--------------------

def main ():
    """The main program"""

    _initialize()
    argumentList = _CommandLineOptions.read()
    _CommandLineOptions.checkArguments(argumentList)

    _process(argumentList.soundFontFilePath,
             argumentList.configurationFilePath,
             argumentList.loggingFilePath,
             argumentList.outputIsPrettyPrinted)

    _finalize()

#--------------------
#--------------------

if __name__ == "__main__":
    main()



# TTBase - provides several elementary functions like conditional
#          expressions
#
# author: Dr. Thomas Tensi, 2014

#====================

import sys

from basemodules.simpletypes import Boolean, IntegerList, Object, Real, \
                                    String

#====================

missingValue = "@!XYZZY"
infinity = 1e99

isStdPython = (sys.implementation.name == "cpython")
isMicroPython = (sys.implementation.name == "micropython")

#====================

def iif (condition : Boolean, trueValue : Object,
         falseValue : Object) -> Object:
    """Emulates conditional expressions with full value evaluation."""

    if condition:
        result = trueValue
    else:
        result = falseValue

    return result

#--------------------

def iif2 (condition1 : Boolean, trueValue1 : Object,
          condition2 : Boolean, trueValue2 : Object,
          falseValue2 : Object) -> Object:
    """Emulates a sequence of conditional expressions with full
       condition and value evaluation."""

    return iif(condition1, trueValue1,
               iif(condition2, trueValue2, falseValue2))

#--------------------

def iif3 (condition1 : Boolean, trueValue1 : Object,
          condition2 : Boolean, trueValue2 : Object,
          condition3 : Boolean, trueValue3 : Object,
          falseValue3 : Object) -> Object:
    """Emulates a sequence of conditional expressions with full
       condition and value evaluation."""

    return iif(condition1, trueValue1,
               iif2(condition2, trueValue2,
                    condition3, trueValue3, falseValue3))

#--------------------

def iif4 (condition1 : Boolean, trueValue1 : Object,
          condition2 : Boolean, trueValue2 : Object,
          condition3 : Boolean, trueValue3 : Object,
          condition4 : Boolean, trueValue4 : Object,
          falseValue4 : Object) -> Object:
    """Emulates a sequence of conditional expressions with full
       condition and value evaluation."""

    return iif(condition1, trueValue1,
               iif3(condition2, trueValue2,
                    condition3, trueValue3,
                    condition4, trueValue4, falseValue4))

#--------------------

def isInRange (x : Object,
               lowBound : Object,
               highBound : Object):
    """Tells whether x lies in the range from <lowBound> to
       <highBound>."""

    return (lowBound <= x <= highBound)

#--------------------

def adaptToRange (x : Object,
                  lowBound : Object,
                  highBound : Object,
                  isCyclic : Boolean = False):
    """Adapts <x> to range [<lowBound>, <highBound>] either by
       clipping at the bounds or <ifCyclic> by shifting periodically"""

    if isInRange(x, lowBound, highBound):
        result = x
    elif not isCyclic:
        result = iif(x < lowBound, lowBound, highBound)
    else:
        intervalLength = highBound - lowBound

        if intervalLength < 1e-4:
            result = lowBound
        else:
            result = x

            while result < lowBound:
                result += intervalLength
            while result > highBound:
                result -= intervalLength

    return result

#--------------------

def intListToHex (currentList : IntegerList) -> String:
    """Returns hex representation of integer <currentList>"""

    return "".join(map(lambda x: ("%02X" % x), currentList))

#====================

class MyRandom:
    """This module provides a simple but reproducible random
       generator."""

    value = None

    #--------------------

    @classmethod
    def initialize (cls):
        """Initializes seed to predefined value"""
        cls.value = 0.123456789

    #--------------------

    @classmethod
    def random (cls) -> Real:
        """Returns a random number in interval [0, 1["""

        cls.value *= 997.0
        cls.value -= int(cls.value)
        return cls.value



# -*- coding: utf-8 -*- 
# configurationfile - provides reading from a configuration file containing
#                     comments and assignment to variables
#
# author: Dr. Thomas Tensi, 2014-04

#====================

import io
import re

from basemodules.operatingsystem import OperatingSystem
from basemodules.simpleassertion import Assertion
from basemodules.simplelogging import Logging
from basemodules.simpletypes import Dictionary, List, Natural, Object, \
                                    ObjectSet, Map, String, StringList, \
                                    StringMap, StringSet, Tuple
from basemodules.typesupport import isString
from basemodules.ttbase import iif, isStdPython, missingValue

#====================

def _reprOfStringToValueMap (stringMap : Map) -> String:
    """String representation for a string to value map <stringMap>"""

    entrySeparator = u"§"
    entryTemplate = "%s: %s"
    keyList = sorted(list(stringMap.keys()))
    result = ""
    
    for key in keyList:
        value = stringMap[key] 
        result += (iif(result == "", "", entrySeparator)
                   + entryTemplate % (key, value))
    
    result = "{" + result + "}";
    return result
        
#====================

class _Token:
    """Simple token within table definition string parser"""

    Kind_number        = "number"
    Kind_string        = "string"
    Kind_operator      = "operator"
    Kind_realNumber    = "real"
    Kind_integerNumber = "integer"

    #--------------------

    def __init__ (self,
                  start : Natural, text : String,
                  kind : String, value : Object):
        """Initializes token with start position, token text, token kind and
           value"""

        self.start = start
        self.text  = text
        self.kind  = kind
        self.value = value
        
    #--------------------

    def __repr__ (self) -> String:
        """String representation for token <self>"""

        st = ("_Token(%r, '%s', %s, %r)"
              % (self.start, self.text, self.kind, self.value))
        return st
        
#====================

if isStdPython:
    _TokenList = List[_Token]
else:
    _TokenList = List # List[_Token]

#====================

class ConfigurationFile:
    """Provides services for reading a configuration file with key -
       value assignments.  The parsing process calculates a map from
       name to value where the values may be booleans, integers,
       reals or strings."""

    _importCommandName = "INCLUDE"
    _trueBooleanValueNames = ["TRUE", "WAHR"]
    _validBooleanValueNames = _trueBooleanValueNames + ["FALSE", "FALSCH"]
    _commentMarker = "--"
    _continuationMarker = "\\"
    _realRegExp = re.compile(r"^[+\-]?[0-9]+\.[0-9]*$")
    _integerRegExp = re.compile(r"^[+\-]?[0-9]+$")
    _hexIntegerRegExp = re.compile(r"^0[xX][0-9A-Fa-f]+$")
    _keyValueRegExp = re.compile(r"^(\w+)\s*=\s*(.*)$")
    _whiteSpaceCharRegExp = re.compile(r"^\s$")
    _identifierCharRegExp = re.compile(r"[A-Za-z_0-9]")
    _identifierFirstCharRegExp = re.compile(r"[A-Za-z_]")
    _identifierLineRegExp = re.compile(r"\s*"
                                       + r"[A-Za-z_][A-Za-z0-9_]*"
                                       + r"\s*=")
    _escapeCharacter = "\\"
    _doubleQuoteCharacter = '"'

    _searchPathList = ["."]

    # error messages
    _errorMsg_badOpeningCharacter = "expected either { or ["
    
    #--------------------
    # LOCAL FEATURES
    #--------------------

    @classmethod
    def _adaptConfigurationValue (cls, value : String) -> Object:
        """Takes string <value> and constructs either a boolean, a numeric
           value or a sanitized string."""

        Logging.trace(">>: %r", value)
        uppercasedValue = value.upper()

        if uppercasedValue in cls._validBooleanValueNames:
            result = (uppercasedValue in cls._trueBooleanValueNames)
        elif cls._integerRegExp.match(value):
            result = int(value)
        elif cls._hexIntegerRegExp.match(value):
            result = int(value, 16)
        elif cls._realRegExp.match(value):
            result = float(value)
        else:
            result = value
            
        Logging.trace("<<: %r", result)
        return result
        
    #--------------------

    @classmethod
    def _combineFragmentedString (cls, st : String) -> String:
        """Combines - possibly fragmented - external representation of a
           string given by <st> into a sanitized string."""

        Logging.trace(">>: %r", st)

        ParseState_inLimbo   = 0
        ParseState_inOther   = 1
        ParseState_inString  = 2
        ParseState_inLiteral = 3
        ParseState_inEscape  = 4

        parseState = ParseState_inLimbo
        result = ""

        for ch in st:
            # process finite state automaton with three states based
            # on next character in string
            # Logging.trace("--: (%d) character: %r", parseState, ch)

            if parseState == ParseState_inLimbo:
                if ch == cls._doubleQuoteCharacter:
                    parseState = ParseState_inString
                elif not cls._whiteSpaceCharRegExp.search(ch):
                    parseState = ParseState_inLiteral
                    result += ch
            elif parseState == ParseState_inString:
                if ch == cls._doubleQuoteCharacter:
                    parseState = ParseState_inLimbo
                else:
                    result += ch
                    parseState = iif(ch == cls._escapeCharacter,
                                     ParseState_inEscape, parseState)
            elif parseState == ParseState_inLiteral:
                result += ch
                if cls._whiteSpaceCharRegExp.search(ch):
                    parseState = ParseState_inLimbo
            elif parseState == ParseState_inEscape:
                result += ch
                parseState = ParseState_inString
            else:
                Assertion.check(False,
                                "bad parse state - %s" % parseState)

        Logging.trace("<<: %r", result)
        return result

    #--------------------

    def _expandVariables (self, st : String) -> String:
        """Expands all variables embedded in <st>."""

        Logging.trace(">>: %r", st)
        cls = self.__class__

        # collect identifiers embedded in value and replace them by
        # their value
        ParseState_inLimbo      = 0
        ParseState_inString     = 1
        ParseState_inEscape     = 2
        ParseState_inIdentifier = 3
        parseStateToString = { 0 : "-", 1 : "S",
                               2 : cls._escapeCharacter, 3 : "I" }

        parseState = ParseState_inLimbo
        result = ""
        identifier = ""
        fsaTrace = ""

        for ch in st:
            # process finite state automaton with three states based
            # on next character in string
            fsaTrace += (iif(fsaTrace == "", "", " ")
                         + "[%s] %s" % (parseStateToString[parseState], ch))

            if parseState == ParseState_inLimbo:
                if cls._identifierFirstCharRegExp.search(ch):
                    identifier = ch
                    parseState = ParseState_inIdentifier
                else:
                    result += ch
                    if ch == cls._doubleQuoteCharacter:
                        parseState = ParseState_inString
            elif parseState == ParseState_inString:
                result += ch
                if ch == cls._doubleQuoteCharacter:
                    parseState = ParseState_inLimbo
                elif ch == cls._escapeCharacter:
                    parseState = ParseState_inEscape
            elif parseState == ParseState_inEscape:
                result += ch
                parseState = ParseState_inString
            elif parseState == ParseState_inIdentifier:
                if cls._identifierCharRegExp.search(ch):
                    identifier += ch
                else:
                    identifierValue = self._findIdentifierValue(identifier)
                    result += identifierValue
                    result += ch
                    parseState = iif(ch == cls._doubleQuoteCharacter,
                                     ParseState_inString, ParseState_inLimbo)

        if parseState == ParseState_inIdentifier:
            identifierValue = self._findIdentifierValue(identifier)
            result += identifierValue
            
        Logging.trace("--: accumulatedFSATrace = %s", fsaTrace)
        Logging.trace("<<: %r", result)
        return result

    #--------------------

    def _findIdentifierValue (self, identifier : String) -> String:
        """Returns string representation of associated identifier value
           for <identifier>; if not found in current key to value map, the
           identifier itself is returned"""

        Logging.trace(">>: %s", identifier)
        cls = self.__class__

        if identifier not in self._keyToValueMap:
            # leave identifier as is (it might be some value name like
            # wahr or false
            Logging.traceError("no expansion found")
            result = identifier
        else:
            result = self._keyToValueMap[identifier]

            if not isString(result):
                result = repr(result)
            else:
                result = (cls._doubleQuoteCharacter + result
                          + cls._doubleQuoteCharacter)

        Logging.trace("<<: expanded %s into %r", identifier, result)
        return result

    #--------------------

    def _lookupFileName (self,
                         enclosingDirectoryName : String,
                         originalFileName : String) -> String:
        """Returns file name in search paths based on <originalFileName>"""

        Logging.trace(">>: directory = %r, file = %r",
                      enclosingDirectoryName, originalFileName)

        cls = self.__class__
        result = None
        separator = OperatingSystem.pathSeparator
        simpleFileName = OperatingSystem.basename(originalFileName)
        searchPathList = list(cls._searchPathList)
        searchPathList.append(enclosingDirectoryName)

        for directoryName in searchPathList:
            fileName = iif(directoryName == ".", originalFileName,
                           directoryName + separator + simpleFileName)
            isFound = OperatingSystem.hasFile(fileName)
            Logging.trace("--: %r -> found = %r", fileName, isFound)

            if isFound:
                result = fileName
                break

        Logging.trace("<<: %r", result)
        return result
    
    #--------------------

    @classmethod
    def _mergeContinuationLines (cls, lineList : StringList):
        """Merges continuation lines in <lineList> into single cumulated line
           and replaces continuations by empty lines (to preserve line
           numbers); the continuation logic is simple: a new
           identifier definition line (starting with an identifier and
           an equals sign) starts a new logical line unconditionally,
           otherwise physical lines are collected and combined with
           the previous logical line; embedded comment lines are
           skipped and an empty (whitespace only) physical line stops
           the collection process (unless followed by a continuation
           character)"""

        Logging.trace(">>")

        cumulatedLine = ""
        markerLength = len(cls._continuationMarker)
        lineListLength = len(lineList)

        for i, originalLine in enumerate(lineList):
            currentLine  = originalLine.strip()
            lineList[i] = ""

            if currentLine.endswith(cls._continuationMarker):
                # strip off obsolete continuation marker
                currentLine = currentLine[:-markerLength].rstrip()

            if cls._identifierLineRegExp.match(currentLine):
                # this is a new definition

                if cumulatedLine > "":
                    lineList[i - 1] = cumulatedLine

                cumulatedLine = currentLine
                loggingFormat = "--: new definition %d (%r)"
            elif currentLine.startswith(cls._commentMarker):
                # skip comment
                loggingFormat = "--: skipped comment %d (%r)"
            elif originalLine == "":
                if cumulatedLine == "":
                    loggingFormat = "--: empty line %d (%r)"
                else:
                    lineList[i - 1] = cumulatedLine
                    loggingFormat = \
                        "--: empty line ended previous definition %d (%r)"

                cumulatedLine = ""
            else:
                # this is not an empty line and it does not start with a
                # definition sequence or an import
                cumulatedLine += " " + currentLine
                loggingFormat = "--: collected continuation %d (%r)"

            Logging.trace(loggingFormat, i+1, currentLine)

        if cumulatedLine > "":
            lineList[-1] = cumulatedLine
            Logging.trace("--: final line (%s)", currentLine);

        Logging.trace("<<")

    #--------------------

    @classmethod
    def _mustHave (cls,
                   token : _Token, kindSet : StringSet,
                   valueSet : ObjectSet = None):
        """Ensures that <token> is of a kind in <kindSet>; if
           <valueSet> is not None, token value is also checked"""

        Logging.trace(">>: token = %r, kindSet = %r, valueSet = %r",
                      token, kindSet, valueSet)

        errorPosition, errorMessage = -1, ""

        if token.kind not in kindSet:
            errorPosition = token.start
            errorMessage  = ("expected kind from %r, found %s"
                             % (kindSet, token.kind))
        elif valueSet is not None and token.value not in valueSet:
            errorPosition = token.start
            errorMessage  = ("expected value from %r, found %r"
                             % (valueSet, token.value))

        result = (errorPosition, errorMessage)
        Logging.trace("<<: %r", result)
        return result

    #--------------------

    def _parseConfiguration (self, lineList : StringList):
        """Parses configuration file data given by <lineList> and updates
           key to value and key to string value map."""

        Logging.trace(">>")

        cls = self.__class__
        cls._mergeContinuationLines(lineList)

        for i, currentLine in enumerate(lineList):
            lineNumber = i + 1

            # remove leading and trailing white space from line
            currentLine = currentLine.strip()
            Logging.trace("--: (%d) %s", i+1, currentLine)

            if (currentLine == ""
                or currentLine.startswith(cls._commentMarker)):
                # this is an empty line or comment line => skip it
                pass
            else:
                match = cls._keyValueRegExp.search(currentLine)

                if not match:
                    Logging.traceError("bad line %d without key-value-pair",
                                       lineNumber)
                else:
                    key = match.group(1)
                    value = match.group(2)
                    value = self._expandVariables(value)
                    value = cls._combineFragmentedString(value)
                    self._keyToStringValueMap[key] = value
                    Logging.trace("--: string value %r -> %r", key, value)
                    value = cls._adaptConfigurationValue(value)
                    self._keyToValueMap[key] = value
                    Logging.trace("--: adapted value %r -> %r", key, value)

        Logging.trace("<<: %r", self._keyToValueMap)

    #--------------------

    def _readFile (self,
                   directoryName : String,
                   fileName : String,
                   lineList : StringList,
                   visitedFileNameSet : StringSet):
        """Appends lines of configuration file with <fileName> to <lineList>
           with leading and trailing whitespace stripped; also handles
           embedded imports of files (relative to <directoryName>;
           <visitedFileNameSet> tells which files have already been
           visited"""

        Logging.trace(">>: fileName = %r, directory = %r, visitedFiles = %r",
                      fileName, directoryName, visitedFileNameSet)

        cls = self.__class__
        errorMessage = ""
        isOkay = True

        originalFileName = fileName
        fileName = self._lookupFileName(directoryName, originalFileName)

        if fileName is None:
            errorMessage = "cannot find %r" % originalFileName
            isOkay = False
        elif fileName in visitedFileNameSet:
            Logging.trace("--: file already included %r", originalFileName)
        else:
            visitedFileNameSet.add(fileName)
            directoryName = OperatingSystem.dirname(fileName)

            with io.open(fileName, "rt",
                         encoding="utf-8") as configurationFile:
                configFileLineList = configurationFile.readlines()

                for currentLine in configFileLineList:
                    currentLine = currentLine.strip()
                    isImportLine = \
                        currentLine.startswith(cls._importCommandName)

                    if isImportLine:
                        importedFileName = currentLine.split('"')[1]
                        currentLine = cls._commentMarker + " " + currentLine

                    lineList.append(currentLine)

                    if isImportLine:
                        isAbsolutePath = (importedFileName.startswith("/")
                                          or importedFileName.startswith("\\")
                                          or importedFileName[1] == ":")

                        # if isAbsolutePath:
                        #     directoryPrefix = ""
                        # else:
                        #     directoryName = OperatingSystem.dirname(fileName)
                        #     directoryPrefix = iif(directoryName == ".", "",
                        #                           directoryName
                        #                           + iif(directoryName > "",
                        #                                 "/", ""))

                        #importedFileName = directoryPrefix + importedFileName
                        Logging.trace("--: IMPORT %r", importedFileName)

                        isOkay = self._readFile(directoryName,
                                                importedFileName, lineList,
                                                visitedFileNameSet)
                        if not isOkay:
                            Logging.traceError("import failed for %r in %r",
                                               importedFileName,
                                               cls._searchPathList)
                            isOkay = False
                            break

        Logging.trace("<<: %r, %r", isOkay, errorMessage)
        return isOkay
            
    #--------------------
    # EXPORTED FEATURES
    #--------------------

    @classmethod
    def setSearchPaths (cls, searchPathList : StringList):
        """Sets list of search paths to <searchPathList>."""

        Logging.trace(">>: %r", searchPathList)
        cls._searchPathList = ["."] + searchPathList
        Logging.trace("<<")

    #--------------------

    def __init__ (self, fileName : String):
        """Parses configuration file given by <fileName> and sets
           internal key to value map."""

        Logging.trace(">>: %r", fileName)

        self._keyToValueMap = {}
        self._keyToStringValueMap = {}
        visitedFileNameSet = set()
        lineList = []
        isOkay = self._readFile("", fileName, lineList, visitedFileNameSet)
        self._parseConfiguration(lineList)

        Logging.trace("<<: %s",
                      _reprOfStringToValueMap(self._keyToValueMap))

    #--------------------

    def asStringMap (self) -> StringMap:
        """Returns mapping from all keys in configuration file to their
           effective values"""

        Logging.trace(">>")
        result = dict(self._keyToValueMap)
        Logging.trace("<<: %r", result)
        return result

    #--------------------

    def asDictionary (self) -> Dictionary:
        """Returns mapping from all keys in configuration file to their
           string values as found in the file"""

        Logging.trace(">>")
        result = dict(self._keyToStringValueMap)
        Logging.trace("<<: %r", result)
        return result

    #--------------------

    def keySet (self) -> StringSet:
        """Returns set of all keys in configuration file"""

        Logging.trace(">>")
        result = set(self._keyToValueMap.keys())
        Logging.trace("<<: %r", result)
        return result

    #--------------------

    def value (self,
               key : String,
               defaultValue : Object = missingValue) -> Object:
        """Returns value for <key> in configuration file; if
           <defaultValue> is missing, an error message is logged when
           there is no associated value, otherwise <defaultValue> is
           returned for a missing entry"""

        Logging.trace(">>: key = %s, defaultValue = %r",
                      key, defaultValue)

        isMandatory = (defaultValue == missingValue)
        result = None

        if key in self._keyToValueMap:
            result = self._keyToValueMap[key]
            if isString(result):
                result = result.replace("\\\"", "\"")
        elif isMandatory:
            Logging.traceError("cannot find value for %s", key)
        else:
            result = defaultValue

        Logging.trace("<<: %s", result)
        return result



# WaveFile - a audio wave file with several channels
#
# author: Dr. Thomas Tensi
# version: 2022-07

#====================
# IMPORTS
#====================

import array
import wave

from basemodules.simpleassertion import Assertion
from basemodules.simplelogging import Logging
from basemodules.simpletypes import List, Natural, Real, RealList, \
                                    String, Tuple
from basemodules.ttbase import iif, iif2

#====================

_WaveDataBuffer = List[RealList]

#====================

class WaveFile:
    """Represents a WAV audio file"""

    #--------------------

    def __init__ (self,
                  name : String):
        """Sets a WAV file with <name>"""

        Logging.trace(">>: %s", name)

        self._fileName = name

        Logging.trace("<<: %s", self)

    #--------------------

    def __repr__ (self) -> String:
        """Returns string representation of wave file"""

        clsName = self.__class__.__name__
        template = "%s(fileName = %s)"
        return template % (clsName, self._fileName)
    
    #--------------------

    def readParams (self) -> Tuple:
        """Returns tuple of channel count, data point width, data
           point rate and data point count"""

        Logging.trace(">>: %s", self)

        audioFile = wave.open(self._fileName, "rb")
        channelCount, dataPointWidth, dataPointRate, dataPointCount, _, _ = \
            audioFile.getparams()
        audioFile.close()
        result = (channelCount, dataPointWidth, dataPointRate, dataPointCount)

        Logging.trace("<<: %r", result)
        return result

    #--------------------

    def readData (self) -> _WaveDataBuffer:
        Logging.trace(">>: %s", self)

        audioFile = wave.open(self._fileName, "rb")
        channelCount, dataPointWidth, dataPointRate, dataPointCount, _, _ = \
            audioFile.getparams()
        byteList = audioFile.readframes(dataPointCount)
        audioFile.close()

        byteCount = len(byteList)
        byteCountPerFrame = channelCount * dataPointWidth
        scalingFactor = 1.0 / (1 << (dataPointWidth * 8 - 1))
        result = []

        for i in range(channelCount):
            byteListSlice = lambda x: byteList[x:x + dataPointWidth]
            intValueList = \
                [ int.from_bytes(byteListSlice(offset),
                                 byteorder = 'little',
                                 signed = True)
                  for offset in range(i * dataPointWidth, byteCount,
                                      byteCountPerFrame) ]
            dataPointList = [ scalingFactor * value
                              for value in intValueList ]
            result.append(dataPointList)

        Logging.trace("<<")
        return result

    #--------------------

    def write (self,
               channelCount : Natural,
               dataPointCount : Natural,
               waveDataBuffer : _WaveDataBuffer,
               dataPointRate : Real,
               dataPointWidth : Natural,
               dataPointKind : String):
        """Writes WAV file from <waveDataBuffer> using <channelCount>
           channels with each <dataPointCount> data points, a rate of
           <dataPointRate>, a sample width of <dataPointWidth> and the
           kind (int, float) given by <dataPointKind>"""

        Assertion.pre(channelCount > 0,    "channel count must be positive")
        Assertion.pre(dataPointCount > 0,  "data point count must be positive")
        Assertion.pre(dataPointRate > 0.0, "data point rate must be positive")
        Assertion.pre(dataPointWidth > 0,  "data point with must be positive")

        Logging.trace(">>: waveFile = %s,"
                      + " channelCount = %d, dataPointCount = %d,"
                      + " dataPointRate = %s, dataPointWidth = %d,"
                      + " dataPointKind = %s)",
                      self, channelCount, dataPointCount,
                      dataPointRate, dataPointWidth, dataPointKind)

        if dataPointKind == "f":
            Logging.traceError("float not supported")
        else:
            scalingFactor = float(1 << (dataPointWidth * 8 - 1))
            adaptationProc = \
                lambda x: int(scalingFactor
                              * (0.99999999 if x >= 1.0
                                 else -1.0 if x < -1.0
                                 else x))
            Logging.trace("--: type adaptation start")

            for channelIndex in range(channelCount):
                waveDataBuffer[channelIndex] = \
                    list(map(adaptationProc, waveDataBuffer[channelIndex]))

            Logging.trace("--: type adaptation end")

            # make a single data point buffer for quicker write
            totalDataPointCount = dataPointCount * channelCount

            if channelCount == 1:
                rawDataPointList = waveDataBuffer[0]
            else:
                referenceDataPoint = waveDataBuffer[0][0]
                Logging.trace("--: before buffer allocation:"
                              + " dataPointCount = %d,"
                              + " totalDataPointCount = %d",
                              dataPointCount, totalDataPointCount)

                # weave the samples into raw data point list
                rawDataPointList = [ referenceDataPoint ] * totalDataPointCount
                Logging.trace("--: before weaving")

                for channelIndex in range(channelCount):
                    rawDataPointList[channelIndex::channelCount] = \
                        waveDataBuffer[channelIndex]

            byteCount = totalDataPointCount * dataPointWidth
            Logging.trace("--: copy start for %d bytes", byteCount)
            byteSequence = [ value.to_bytes(dataPointWidth, 'little',
                                            signed = True)
                             for value in rawDataPointList ]
            byteArray = b''.join(byteSequence)

            with wave.open(self._fileName, "wb") as audioFile:
                audioFile.setnchannels(channelCount)
                audioFile.setsampwidth(dataPointWidth)
                audioFile.setframerate(dataPointRate)
                audioFile.setnframes(dataPointCount)
                Logging.trace("--: writing frames to file")
                audioFile.writeframesraw(byteArray)

        Logging.trace("<<")




# simpletypes - provide the internal type names like String, Real, ...
#
# author: Dr. Thomas Tensi, 2014

#====================

import sys

isMicroPython = (sys.implementation.name == "micropython")

if isMicroPython:
    import basemodules.typing as typing
else:
    import typing

#====================

Class    = type
ClassVar = typing.ClassVar
DataType = type

# primitive types
Bit       = int
Boolean   = bool
Byte      = int
Character = str
Integer   = int
Natural   = int
Object    = typing.Any
Positive  = int
Real      = float
String    = str

# list types
List        = typing.Sequence
ObjectList  = List
Tuple       = typing.Tuple
Pair        = Tuple

if isMicroPython:
    BitList     = List #List[Bit]
    ByteList    = List #List[Byte]
    IntegerList = List #List[Integer]
    NaturalList = List #List[Natural]
    RealList    = List #List[Real]
    StringList  = List #List[String]
    TupleList   = List #List[Tuple]
else:
    BitList     = List[Bit]
    ByteList    = bytearray
    IntegerList = List[Integer]
    NaturalList = List[Natural]
    RealList    = List[Real]
    StringList  = List[String]
    TupleList   = List[Tuple]

# set types
Set         = typing.Set
ObjectSet   = Set

if isMicroPython:
    StringSet   = Set #Set[String]
else:
    StringSet   = Set[String]

# mapping types
Map        = typing.Mapping

if isMicroPython:
    Dictionary = Map #Map[String, String]
    StringMap  = Map #Map[String, Object]
else:
    Dictionary = Map[String, String]
    StringMap  = Map[String, Object]

# function types
Callable = typing.Callable



# validitychecker - allows checking of validity of values (typically
#                   of input parameters from the command line)
#
# author: Dr. Thomas Tensi, 2014-04

#====================

import os
import re
import sys

import basemodules.typesupport as typesupport
from basemodules.operatingsystem import OperatingSystem
from basemodules.simplelogging import Logging
from basemodules.simpletypes import Boolean, Object, String
from basemodules.ttbase import iif, iif2, isStdPython

if isStdPython:
    import subprocess

#====================

class ValidityChecker:
    """Provides checking of validity of values (typically command line
       parameters). Typically assumes that a check failure is fatal."""

    #--------------------
    # LOCAL FEATURES
    #--------------------

    _rangeKindToTemplateMap = {
        ""     : "%s",
        ">0"   : "a positive %s",
        ">=0"  : "a non-negative %s"
    }

    #--------------------

    @classmethod
    def _checkTemplate (cls,
                        prefix : String,
                        typeName : String,
                        valueName : String,
                        value : Object,
                        failureCausesExit : Boolean):
        result = (prefix + ": checking %s for being %s (%r) "
                  + iif(failureCausesExit, "with", "without")
                  + " failure on exit")
        result = result % (valueName, typeName, value)
        # remove any template characters in result string
        result = result.replace("%", "§")
        return result

    #--------------------

    @classmethod
    def _checkForType (cls,
                       kind : String,
                       kindName : String,
                       valueName : String,
                       value : Object,
                       failureCausesExit : Boolean = True) -> Boolean:
        Logging.trace(cls._checkTemplate(">>", kindName, valueName, value,
                                         failureCausesExit))

        message = "%s must be %s: %r" % (valueName, kindName, value)
        isOkay = isinstance(value, kind)
        cls.isValid(isOkay, message, failureCausesExit)

        Logging.trace("<<: %r", isOkay)
        return isOkay

    #--------------------

    @classmethod
    def _constructErrorMessage (cls,
                                template : String,
                                value : Object,
                                valueName : String) -> String:
        """Makes an error message from <template> and <value>
           depending on whether <valueName> is given or not."""

        if valueName is None:
            result = template % (value, "")
        else:
            result = template % (value, valueName)

        return result

    #--------------------
    # EXPORTED FEATURES
    #--------------------

    @classmethod
    def isOfKind (cls,
                  value : Object,
                  valueName : String,
                  kind : String,
                  failureCausesExit : Boolean = True) -> Boolean:
        """Checks whether <value> named <valueName> has <kind>;
           otherwise gives <message> and exits program when
           <failureCausesExit> is set"""

        Logging.trace(">>: value = %r, valueName = %s, kind = %r,"
                      + " failureCausesExit = %s",
                      value, valueName, kind, failureCausesExit)

        if kind not in _kindToCheckProcMap:
            isOkay = False
        else:
            isOkay = _kindToCheckProcMap[kind](value, valueName,
                                               failureCausesExit)

        Logging.trace("<<: %r", isOkay)
        return isOkay
    
    #--------------------

    @classmethod
    def isBoolean (cls,
                   value : Object,
                   valueName : String,
                   failureCausesExit : Boolean = True) -> Boolean:
        """Checks whether <value> of variable given by <valueName> is
           a boolean, otherwise gives <message> and exits program when
           <failureCausesExit> is set."""

        Logging.trace(">>: value = %r, valueName = %s,"
                      + " failureCausesExit = %s",
                      value, valueName, failureCausesExit)

        isOkay = cls._checkForType(bool, "a boolean", valueName, value,
                                   failureCausesExit)

        Logging.trace("<<: %r", isOkay)
        return isOkay

    #--------------------

    @classmethod
    def isBooleanString (cls,
                         value : String,
                         valueName : String,
                         failureCausesExit : Boolean = True) -> Boolean:
        """Checks whether <value> of variable given by <valueName> is
           a boolean string, otherwise gives <message> and exits
           program when <failureCausesExit> is set."""

        Logging.trace(">>: value = %r, valueName = %s,"
                      + " failureCausesExit = %s",
                      value, valueName, failureCausesExit)

        value = str(value)
        isOkay = value.lower() in ["true", "false"]
        message = "%s must be a boolean string: %r" % (valueName, value)
        cls.isValid(isOkay, message, failureCausesExit)

        Logging.trace("<<: %r", isOkay)
        return isOkay

    #--------------------

    @classmethod
    def isDirectory (cls,
                     pathName : String,
                     valueName : String = None,
                     failureCausesExit : Boolean = True) -> Boolean:
        """Checks whether directory given by <pathName> is readable,
           otherwise gives <message> and exits program when
           <failureCausesExit> is set."""

        Logging.trace(cls._checkTemplate(">>", "a directory", valueName,
                                         pathName, failureCausesExit))

        template = "Directory %r (%s) is not readable."
        message = cls._constructErrorMessage(template, pathName, valueName)
        isOkay = os.path.isdir(pathName)
        cls.isValid(isOkay, message, failureCausesExit)

        Logging.trace("<<: %r", isOkay)
        return isOkay

    #--------------------

    @classmethod
    def isExecutableCommand (cls,
                             commandName : String,
                             valueName : String = None,
                             failureCausesExit : Boolean = True) -> Boolean:
        """Checks whether executable given by <commandName> can be
           executed, otherwise gives <message> and exits program when
           <failureCausesExit> is set."""

        Logging.trace(cls._checkTemplate(">>", "an executable command",
                                         valueName, commandName,
                                         failureCausesExit))

        template = "Command %r (%s) cannot be executed."
        message = cls._constructErrorMessage(template, commandName,
                                             valueName)

        if not isStdPython:
            isOkay = False
        else:
            try:
                subprocess.run(commandName,
                               stdout=subprocess.DEVNULL,
                               stderr=subprocess.DEVNULL)
                isOkay = True
            except:
                isOkay = False

            if not isOkay:
                try:
                    subprocess.call(commandName)
                    isOkay = True
                except:
                    pass

        cls.isValid(isOkay, message, failureCausesExit)

        Logging.trace("<<: %r", isOkay)
        return isOkay

    #--------------------

    @classmethod
    def isInteger (cls,
                   value : String,
                   valueName : String,
                   failureCausesExit : Boolean = True) -> Boolean:
        """Checks whether <value> of variable given by <valueName> is
           an integer or long integer, otherwise exits program with a
           message."""

        Logging.trace(cls._checkTemplate(">>", "an integer", valueName,
                                         value, failureCausesExit))

        message = "%s must be an integer: %r" % (valueName, value)
        isOkay = typesupport.isInteger(value)
        cls.isValid(isOkay, message, failureCausesExit)

        Logging.trace("<<: %r", isOkay)
        return isOkay

    #--------------------

    @classmethod
    def isList (cls,
                value : String,
                valueName : String,
                failureCausesExit : Boolean = True) -> Boolean:
        """Checks whether <value> of variable given by <valueName> is
           a list, otherwise gives <message> and exits program when
           <failureCausesExit> is set."""

        Logging.trace(">>: value = %r, valueName = %s,"
                      + " failureCausesExit = %s",
                      value, valueName, failureCausesExit)

        isOkay = cls._checkForType(list, "a list", valueName, value,
                                   failureCausesExit)

        Logging.trace("<<: %r", isOkay)
        return isOkay

    #--------------------

    @classmethod
    def isMap (cls,
               value : String,
               valueName : String,
               failureCausesExit : Boolean = True) -> Boolean:
        """Checks whether <value> of variable given by <valueName> is
           a map, otherwise gives <message> and exits program when
           <failureCausesExit> is set."""

        Logging.trace(">>: value = %r, valueName = %s,"
                      + " failureCausesExit = %s",
                      value, valueName, failureCausesExit)

        isOkay = cls._checkForType(dict, "a map", valueName, value,
                                   failureCausesExit)

        Logging.trace("<<: %r", isOkay)
        return isOkay

    #--------------------

    @classmethod
    def isNatural (cls,
                   value : String,
                   valueName : String,
                   zeroIsIncluded : Boolean = True,
                   failureCausesExit : Boolean = True) -> Boolean:
        """Checks whether <value> of variable given by <valueName> is
           a positive integer or long integer, otherwise gives
           <message> and exits program when <failureCausesExit> is
           set.  When <zeroIsIncluded> is set, also zero is
           acceptable."""

        kindName = "a " + iif(zeroIsIncluded, "", "positive ") + "natural"
        Logging.trace(cls._checkTemplate(">>", kindName, valueName,
                                         value, failureCausesExit))
        message = ("%s must be %s: %r" % (valueName, kindName, value))
        isOkay = (typesupport.isInteger(value)
                  and (value > 0 or value == 0 and zeroIsIncluded))
        cls.isValid(isOkay, message, failureCausesExit)

        Logging.trace("<<: %r", isOkay)
        return isOkay

    #--------------------

    @classmethod
    def isNumberString (cls,
                        value : String,
                        valueName : String,
                        realIsAllowed : Boolean,
                        rangeKind : String = "",
                        failureCausesExit : Boolean = True) -> Boolean:
        """Checks whether string <value> with name <valueName> is
           representation of a correct number. <realIsAllowed> tells
           whether non-integer values are okay, <rangeKind> gives an
           boundary condition about the range."""

        Logging.trace(">>: %s = %r (%s), realIsOk = %r, rangeKind = %r,"
                      + " failureCausesExit = %s",
                      valueName, value, type(value), realIsAllowed,
                      rangeKind, failureCausesExit)

        realRegexp   = re.compile(r"^\-?[0-9]+(\.[0-9]*)?$")
        integerRegexp = re.compile(r"^\-?[0-9]+$")

        value = str(value)
        isOkay = (integerRegexp.match(value) is not None)

        if rangeKind not in ("", ">0", ">=0"):
            rangeKind = ""

        if realIsAllowed and not isOkay:
            isOkay = (realRegexp.match(value) is not None)

        if isOkay:
            isOkay = iif2(rangeKind == ">0",  (float(value) > 0),
                          rangeKind == ">=0", (float(value) >= 0),
                          True)

        typeDesignation = (cls._rangeKindToTemplateMap[rangeKind]
                           % iif(realIsAllowed, "real", "integer"))
        errorTemplate = "%s must be " + typeDesignation + " - %r"
        cls.isValid(isOkay, errorTemplate % (valueName, value),
                    failureCausesExit)

        Logging.trace("<<: %r", isOkay)
        return isOkay

    #--------------------

    @classmethod
    def isReadableFile (cls,
                        pathName : String,
                        valueName : String = None,
                        failureCausesExit : Boolean = True) -> Boolean:
        """Checks whether file given by <pathName> is readable,
           otherwise gives <message> and exits program when
           <failureCausesExit> is set."""

        Logging.trace(cls._checkTemplate(">>", "a readableFile",
                                         valueName, pathName,
                                         failureCausesExit))
        template = "File %r (%s) is not readable."
        message = cls._constructErrorMessage(template, pathName, valueName)
        isOkay = False if pathName is None else os.path.isfile(pathName)
        cls.isValid(isOkay, message, failureCausesExit)

        Logging.trace("<<: %r", isOkay)
        return isOkay

    #--------------------

    @classmethod
    def isReal (cls,
                value : String,
                valueName : String,
                failureCausesExit : Boolean = True) -> Boolean:
        """Checks whether <value> of variable given by <valueName> is
           a real number, otherwise gives <message> and exits program
           when <failureCausesExit> is set."""

        Logging.trace(">>: value = %r, valueName = %s,"
                      + " failureCausesExit = %s",
                      value, valueName, failureCausesExit)

        isOkay = cls._checkForType(float, "a real number", valueName,
                                   value, failureCausesExit)

        Logging.trace("<<: %r", isOkay)
        return isOkay

    #--------------------

    @classmethod
    def isString (cls,
                  value : String,
                  valueName : String,
                  failureCausesExit : Boolean = True) -> Boolean:
        """Checks whether <value> of variable given by <valueName> is
           a unicode string, otherwise gives <message> and exits
           program when <failureCausesExit> is set."""

        Logging.trace(cls._checkTemplate(">>", "a (unicode) string",
                                         valueName, value,
                                         failureCausesExit))

        message = "%s must be a string: %r" % (valueName, value)
        isOkay = typesupport.isString(value)
        cls.isValid(isOkay, message, failureCausesExit)

        Logging.trace("<<: %r", isOkay)
        return isOkay

    #--------------------

    @classmethod
    def isStringList (cls,
                      value : String,
                      valueName : String,
                      failureCausesExit : Boolean = True) -> Boolean:
        """Checks whether <value> of variable given by <valueName> is
           a (unicode) string list, otherwise gives <message> and
           exits program when <failureCausesExit> is set."""

        Logging.trace(cls._checkTemplate(">>", "a (unicode) string list",
                                         valueName, value,
                                         failureCausesExit))

        message = "%s must be a string list: %r" % (valueName, value)
        isOkay = typesupport.isStringList(value)
        cls.isValid(isOkay, message, failureCausesExit)

        Logging.trace("<<: %r", isOkay)
        return isOkay

    #--------------------

    @classmethod
    def isStringMap (cls,
                     value : String,
                     valueName : String,
                     failureCausesExit : Boolean = True) -> Boolean:
        """Checks whether <value> of variable given by <valueName> is a
           map from (unicode) strings to other objects, otherwise exits
           program with a message."""

        Logging.trace(cls._checkTemplate(">>", "a (unicode) string map",
                                         valueName, value,
                                         failureCausesExit))
        message = "%s must be a string map: %r" % (valueName, value)
        isOkay = typesupport.isStringMap(value)
        cls.isValid(isOkay, message, failureCausesExit)

        Logging.trace("<<: %r", isOkay)
        return isOkay

    #--------------------

    @classmethod
    def isWritableFile (cls,
                        pathName : String,
                        valueName : String = None,
                        failureCausesExit : Boolean = True) -> Boolean:
        """Checks whether file given by <pathName> is writable,
           otherwise gives <message> and exits program when
           <failureCausesExit> is set."""

        Logging.trace(cls._checkTemplate(">>", "a writable file",
                                         valueName, pathName,
                                         failureCausesExit))

        template = "File %r (%s) cannot be written."
        message = cls._constructErrorMessage(template, pathName, valueName)
        directoryName = ("???" if pathName is None
                         else OperatingSystem.dirname(pathName))
        directoryName = iif(directoryName == "", ".", directoryName)
        Logging.trace("--: dir = %r", directoryName)
        isOkay = os.path.isdir(directoryName)
        cls.isValid(isOkay, message, failureCausesExit)

        Logging.trace("<<: %r", isOkay)
        return isOkay

    #--------------------

    @classmethod
    def isValid (cls,
                 condition : Boolean,
                 message : String,
                 failureCausesExit : Boolean = True):
        """Checks whether <condition> holds, otherwise gives <message>
           and exits program when <failureCausesExit> is set."""

        Logging.trace("--: checking condition (%r),"
                      + " otherwise failure is %r",
                      condition, message)

        if not condition:
            Logging.traceError(message)

            if failureCausesExit:
                sys.exit("ERROR: " + message)

#--------------------

_kindToCheckProcMap = {
    "B"  : ValidityChecker.isBoolean,
    "EX" : ValidityChecker.isExecutableCommand,
    "I"  : ValidityChecker.isInteger,
    "N"  : ValidityChecker.isInteger,
    "PN" : ValidityChecker.isInteger,
    "R"  : ValidityChecker.isReal,
    "RF" : ValidityChecker.isReadableFile,
    "S"  : ValidityChecker.isString,
    "SL" : ValidityChecker.isStringList,
    "SM" : ValidityChecker.isStringMap,
    "WD" : ValidityChecker.isDirectory,
    "WF" : ValidityChecker.isWritableFile
}



# -*- coding:utf-8 -*-
# StringUtil - provides several utility string functions like conversions
#              from strings to maps or lists or tokenization
#
# author: Dr. Thomas Tensi, 2014

#====================

from basemodules.simplelogging import Logging
from basemodules.simpletypes import IntegerList, Map, Natural, Object, \
                                    ObjectList, String, StringList, \
                                    StringMap, Tuple
from basemodules.ttbase import iif, intListToHex

#====================

def adaptToKind (st : String,
                 kind : String) -> Object:
    """Returns converted value for given string <st> with type given by
       <kind>."""

    if kind == "I":
        result = int(st)
    elif kind == "R":
        result = float(st)
    elif kind == "B":
        result = (st.upper() == "TRUE")
    else:
        result = st

    return result

#--------------------

def newlineReplacedString (st : String) -> String:
    """Returns form of <st> where newlines are replaced by hash characters
       (to have multiline strings in a single line)"""

    replacementCharacter = "#"
    result = (st
              .replace("\r\n", replacementCharacter)
              .replace("\n", replacementCharacter)
              .replace("\r", replacementCharacter))
    return result

#--------------------

def shortenedString (st : String,
                     maximumLength : Natural = 20) -> String:
    """Returns <st> shortened to <maximumLength>"""

    if len(st) > maximumLength:
        result = st[:maximumLength - 1] + u"…"
        #result = st[:maximumLength - 3] + "..."
    else:
        result = st

    return result

#--------------------

def stringToIntList (st : String) -> IntegerList:
    """Returns integer list for string <st>"""

    intList = [ ord(st[i]) for i in range(len(st)) ]
    return intList

#--------------------

def stringToHex (st : String) -> String:
    """Returns hex representation of <st>"""

    return intListToHex(stringToIntList(st))

#--------------------

def splitAndStrip (st : String,
                   separator : String) -> StringList:
    """Returns split of <st> by <separator> with each part stripped from
       leading and trailing whitespace"""

    return [x.strip() for x in st.split(separator)]

#--------------------

def splitAt (st : String,
             separator : String) -> Tuple:
    """Returns split of <st> by <separator> at first position; if
       there is none, the third result is a False value"""

    separatorPosition = st.find(separator)
    separatorLength   = len(separator)
    isFound = (separatorPosition >= 0)

    if isFound:
        partA = st[:separatorPosition]
        partB = st[separatorPosition+separatorLength:]
    else:
        partA, partB = (st, "")

    result = (partA, partB, isFound)
    return result

#--------------------

def stripStringQuotes (st : String) -> String:
    """Returns <st> with string quotes removed"""

    for ch in ("'", "\""):
        if st.startswith(ch) and st.endswith(ch):
            st = st[1:-1]
    
    return st

#--------------------

def tokenize (st : String) -> StringList:
    """Returns a list of tokens also taking care of strings"""

    Logging.trace(">>: %r", st)

    whiteSpaceCharacterList = " "
    quoteCharacterList = "\"'"
    escapeCharacter = '\\'

    ParseState_inLimbo  = 0
    ParseState_inString = 1
    ParseState_inEscape = 2
    ParseState_inToken  = 3
    parseStateToString = { 0 : "-", 1 : "S", 2 : escapeCharacter, 3 : "T" }

    parseState = ParseState_inLimbo
    result = []
    token = ""
    fsaTrace = ""

    for ch in st:
        # process finite state automaton with three states based
        # on next character in string
        fsaTrace += (iif(fsaTrace == "", "", " ")
                     + "[%s] %s" % (parseStateToString[parseState], ch))

        if parseState == ParseState_inLimbo:
            if ch in whiteSpaceCharacterList:
                pass
            elif ch in quoteCharacterList:
                endCharacter = ch
                token = ""
                parseState = ParseState_inString
            else:
                token = ch
                parseState = ParseState_inToken
        elif parseState == ParseState_inString:
            if ch == endCharacter:
                result.append(token)
                token = ""
                parseState = ParseState_inLimbo
            else:
                token += ch
                parseState = iif(ch == escapeCharacter, ParseState_inEscape,
                                 parseState)
        elif parseState == ParseState_inEscape:
            result += ch
            parseState = ParseState_inString
        elif parseState == ParseState_inToken:
            if ch in whiteSpaceCharacterList:
                result.append(token)
                token = ""
                parseState = ParseState_inLimbo
            elif ch in quoteCharacterList:
                result.append(token)
                token = ch
                parseState = ParseState_inString
            else:
                token += ch

    if parseState != ParseState_inLimbo:
        result.append(token)

    Logging.trace("--: accumulatedTrace = %r", fsaTrace)
    Logging.trace("<<: %r", result)
    return result

#===============================
# STRING TO CONTAINER CONVERSION
#===============================

_quoteCharacterSet           = set(["\"", "'"])
_structureLeadinCharacterSet = set(["{", "[", "("])

#--------------------

def _deserializeToList (st : String,
                        startPosition : Natural,
                        separator : String) -> StringList:
    """Splits <st> starting from <startPosition> at <separator> into
       list of parts; sublists or submaps are correctly handled and
       embedded into the list; returns end position and resulting list"""

    Logging.trace(">>: separator = %r, st = %s#|#%s",
                  separator, st[:startPosition], st[startPosition:])

    # do a finite state automaton with "{", "[", "(", " ", "\"" and "'"
    # as state changing inputs
    ParseState_beforeElement = 1
    ParseState_inElement     = 2
    ParseState_inString      = 3
    ParseState_afterElement  = 4
    ParseState_afterList     = 5

    currentList = []
    lastPosition = len(st) - 1
    listEndCharacter = iif(st[startPosition] == "[", "]", ")")
    position = startPosition + 1
    parseState = ParseState_beforeElement
    fsaTrace = ""

    while position <= lastPosition:
        ch = st[position]
        fsaTrace += (" [%d] %s" % (parseState, ch))

        if parseState == ParseState_beforeElement:
            if ch == listEndCharacter:
                break
            elif ch == " ":
                pass
            elif ch == separator:
                # unexpected separator => ignore
                pass
            elif ch in _quoteCharacterSet:
                endQuote = ch
                currentElement = ""
                parseState = ParseState_inString
            elif ch in _structureLeadinCharacterSet:
                if ch == "{":
                    position, currentElement = \
                              _deserializeToMap(st, position, separator)
                else:
                    position, currentElement = \
                              _deserializeToList(st, position, separator)

                currentList.append(currentElement)
                parseState = ParseState_afterElement
            else:
                currentElement = ch
                parseState = ParseState_inElement
        elif parseState == ParseState_inElement:
            if ch not in [" ", separator, listEndCharacter]:
                currentElement += ch
            else:
                currentList.append(currentElement)

                if ch == listEndCharacter:
                    break
                else:
                    parseState = iif(ch == " ", ParseState_afterElement,
                                     ParseState_beforeElement)
        elif parseState == ParseState_inString:
            if ch != endQuote:
                currentElement += ch
            else:
                currentList.append(currentElement)
                parseState = ParseState_afterElement
        elif parseState == ParseState_afterElement:
            # ignore everything except a separator or a list end
            if ch == listEndCharacter:
                break
            elif ch == separator:
                parseState = ParseState_beforeElement

        position += 1

    Logging.trace("--: fsa = %s", fsaTrace)
    result = (position, currentList)
    Logging.trace("<<: %r", result)
    return result

#--------------------

def _deserializeToMap (st : String,
                       startPosition : Natural,
                       separator : String) -> StringMap:
    """Splits <st> starting from <startPosition> at <separator> into map
       of key-value-pairs; sublists or submaps as values are correctly
       handled and embedded into the map; returns end position and
       resulting map"""

    Logging.trace(">>: separator = %r, st = %s#|#%s",
                  separator, st[:startPosition], st[startPosition:])

    # do a finite state automaton with "{", "[", "(", " ", "\"" and "'"
    # as state changing inputs
    ParseState_beforeKey     = 1
    ParseState_inKey         = 2
    ParseState_inKeyString   = 3
    ParseState_afterKey      = 4
    ParseState_beforeValue   = 5
    ParseState_inValue       = 6
    ParseState_inValueString = 7
    ParseState_afterValue    = 8
    ParseState_afterMap      = 9

    table = {}

    mapEndCharacter = "}"
    lastPosition = len(st) - 1
    position = startPosition + 1
    parseState = ParseState_beforeKey
    fsaTrace = ""

    while position <= lastPosition:
        ch = st[position]
        fsaTrace += (" [%d] %s" % (parseState, ch))

        if parseState == ParseState_beforeKey:
            currentKey = ""

        if parseState in [ParseState_beforeKey, ParseState_beforeValue]:
            elementIsUnhandled = True

            if ch == mapEndCharacter:
                break
            elif ch == " ":
                pass
            elif ch == separator:
                # unexpected separator => ignore
                pass
            elif ch in _quoteCharacterSet:
                endQuote = ch
                currentElement = ""
                parseState += 2
            elif ch in _structureLeadinCharacterSet:
                if ch == "{":
                    position, currentElement = \
                              _deserializeToMap(st, position, separator)
                else:
                    position, currentElement = \
                              _deserializeToList(st, position, separator)

                parseState += 3
            else:
                currentElement = ch
                parseState += 1
        elif parseState in [ParseState_inKey, ParseState_inValue]:
            if ch not in [" ", separator, ":", mapEndCharacter]:
                currentElement += ch
            else:
                parseState += 2
                position -= iif(ch == " ", 0, 1)
        elif parseState in [ParseState_inKeyString, ParseState_inValueString]:
            if ch != endQuote:
                currentElement += ch
            else:
                parseState += 1
        elif parseState in [ParseState_afterKey, ParseState_afterValue]:
            if elementIsUnhandled:
                elementIsUnhandled = False

                if parseState == ParseState_afterKey:
                    currentKey = currentElement
                else:
                    table[currentKey] = currentElement

            # ignore everything except a separator or a colon
            if ch == mapEndCharacter:
                break
            elif ch == separator:
                parseState = ParseState_beforeKey
            elif ch == ":":
                parseState = ParseState_beforeValue

        position += 1

    Logging.trace("--: fsa = %s", fsaTrace)
    result = (position, table)
    Logging.trace("<<: %r", result)
    return result

#--------------------

def deserializeToList (st : String,
                       separator : String = ",",
                       kind : String = "S") -> ObjectList:
    """Splits <st> with parts separated by <separator> into list with
       all parts having leading and trailing whitespace removed; if
       <kind> is 'I' or 'R' the elements are transformed into ints or
       floats"""

    Logging.trace(">>: separator = %r, kind = %r, st = %s",
                  separator, kind, st)
    st = st.strip()
    st = iif(st.startswith("["), st, "[%s]" % st)
    _, result = _deserializeToList(st, 0, separator)
    result = list(map(lambda x: adaptToKind(x, kind), result))
    Logging.trace("<<: %r", result)
    return result

#--------------------

def deserializeToMap (st : String,
                      separator : String = ",") -> Map:
    """Splits <st> with parts separated by <separator> into map where
       key and value is separated by colons with all parts
       having leading and trailing whitespace removed"""

    Logging.trace(">>: separator = %r, st = %s", separator, st)
    result = {}
    st = st.strip()
    st = iif(st.startswith("{"), st, "{%s}" % st)
    _, result = _deserializeToMap(st, 0, separator)
    Logging.trace("<<: %r", result)
    return result

